import numpy as np

from naeural_core import constants as ct

from PIL import Image



def maybe_prepare_img_payload(sender, dct, keys=['IMG', 'IMG_ORIG'], force_list=False):
  dct['IMG_IN_PAYLOAD'] = False
  for key in keys:
    lst_img = dct.get(key, None)
    
    collected = True
  
    if not isinstance(lst_img, list):
      collected = False
      lst_img = [lst_img]
    
    sender.start_timer('all_img_to_msg')
    if isinstance(lst_img[0], np.ndarray):
      dct[key] = []
      dct[key + '_SIZE'] = []
      for img in lst_img:
        if img is None:
          continue
        if img.dtype != np.uint8:
          if img.dtype == 'float' and 0 <= img.min() <= 1 and 0 <= img.max() <= 1:
            img = img * 255.0
          #endif
          img = img.astype(np.uint8)
        #endif
  
        sender.start_timer('np_image_to_base64')
        try:
          is_original = 'ORIG' in key.upper()
          img_str, h, w = img_to_msg(sender.log, img, orig=is_original)
          dct['IMG_IN_PAYLOAD'] = True
        except Exception as e:
          err_msg = str(e)
          msg = "Error occured in 'IMG' preparation for image: {}. Exception: '{}'".format(
            img.shape, err_msg,
          )
          raise ValueError(msg)
        # end try-except on image to text        
        sender.end_timer('np_image_to_base64')
        
        dct[key].append(img_str)
        dct[key + '_HEIGHT'] = h
        dct[key + '_WIDTH'] = w
        dct[key + '_SIZE'].append([h,w])
      #endfor each image in list
      if (not collected) and (not force_list):
        dct[key] = dct[key][0]
        dct[key + '_SIZE'] = dct[key + '_SIZE'][0]
    #endif if np.ndarray
    sender.end_timer('all_img_to_msg')

  return


# When testing we observed that when resizing with PIL
# there was no need to increase the quality above 60 for images
# with height > 1080. Before, if we wanted to compress a 4K image
# we had to increase the quality to 80
# !!! When compressing a 4K image please be sure that
# the font scale and font thickness for the text on the image
# are both 2 for reading clarity
def img_to_msg(log, img, orig=False):
  # retrieving the original resolution
  h, w, _ = np.shape(img)

  if not orig:
    quality=ct.IMAGE_COMPRESSION.QUALITY
    max_height=ct.IMAGE_COMPRESSION.MAX_HEIGHT
  else:
    quality=95
    max_height=ct.IMAGE_COMPRESSION.ORIG_MAX_HEIGHT

  # resizing image and converting to string
  img_str = log.np_image_to_base64(
    np_image=img,
    quality=quality,
    max_height=max_height
  )

  return img_str, h, w

def msg_to_img(log, msg, h, w):
  # decoding image
  img = log.base64_to_np_image(msg)

  # retrieving the decoded resolution
  curr_h, curr_w, _ = np.shape(img)

  # checking if we need to resize it
  if h > curr_h:
    img = Image.fromarray(img)

    # resizing to original size
    img = img.resize((w, h))

    # converting to np array
    img = np.array(img)

  return img



"""
TODO: Refactor in order to have a abstract video stream than can:
  - resize & cropping
  - stream splitting (virtual stream) so we can spawn multiple streams from same one
  - unify all video stream under this abstract video stream

"""
import numpy as np
import cv2
import traceback
from naeural_core import constants as ct
from time import sleep

from naeural_core.data.base import DataCaptureThread
from naeural_core.data.mixins_libs import _VideoConfigMixin

_CONFIG = {
  **DataCaptureThread.CONFIG,

  "MAX_RETRIES": 2,

  "AMD_TARGET_DPS": 0,
  "SIMULATE_AMD": False,

  "CAP_PROP_BUFFERSIZE": 0,

  "USE_FFMPEG": False,

  "CONFIGURED_H": -1,
  "CONFIGURED_W": -1,


  'VALIDATION_RULES': {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],

    "AMD_TARGET_DPS": {
      "DESCRIPTION": "For AMD architectures (as of 2023-04-28 with no plans to use also on Intel) forces number of grabs per iteration based on the target required DPS. Set this to 3 for example to (hopefully) get 3 FPS on a 15 FPS camera",
      "TYPE": "int"
    }
  },
}


class VideoStreamCv2DataCapture(DataCaptureThread, _VideoConfigMixin):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._capture = None
    self._crt_frame = 0
    self.__configured_size_error = False

    super(VideoStreamCv2DataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    self._metadata.update(
      fps=None,
      frame_h=None,
      frame_w=None,
      frame_count=None,
      frame_current=None,
    )
    return

  @property
  def is_hw_error(self):
    return self.__configured_size_error

  def is_intel(self):
    if self.cfg_simulate_amd:
      result = False
    else:
      proc = self.log.get_processor_platform()
      if proc is not None:
        result = "intel" in proc.lower()
      else:
        result = False
    return result

  def _release(self):
    self._release_capture()
    del self._capture
    return

  def _release_capture(self):
    self._capture.release()
    return

  def _init(self):
    self._force_video_stream_config()
    return

  def _open_stream(self):

    if self.cfg_use_ffmpeg:
      self.P("    Opening capture with CAP_FFMPEG...")
      cap = cv2.VideoCapture(self.cfg_url, cv2.CAP_FFMPEG)
    else:
      cap = cv2.VideoCapture(self.cfg_url)

    if self.cfg_cap_prop_buffersize > 0:
      self.P("    Setting cv2.CAP_PROP_BUFFERSIZE={}".format(self.cfg_cap_prop_buffersize))
      cap.set(cv2.CAP_PROP_BUFFERSIZE, self.cfg_cap_prop_buffersize)
    return cap

  def _on_config_changed(self):
    if self.cfg_url != self.last_url:
      self.has_connection = False  # forces reconnection
      self.P("RTSP URI change detected from {} to {}, reconnecting...".format(
        self.last_url, self.cfg_url,
      )
      )
    return

  def _maybe_reconnect(self):
    if self.has_connection:
      return

    self.last_url = self.cfg_url  # a cache of current url
    self.has_connection = False
    nr_retry = 0
    if self.nr_connection_issues == 0:
      msg = "Connecting to url:"
      color = 'g'
    else:
      msg = "Reconnecting ({}) to url:".format(self.nr_connection_issues)
      color = 'r'
    self.P("{} {} (deq size: {})".format(msg, self.cfg_url, self._deque.maxlen), color=color)
    str_e = None
    while nr_retry <= self.cfg_max_retries:
      self.sleep(1)
      nr_retry += 1
      self.nr_connection_issues += 1
      try:
        if self._capture:
          self._release_capture()
        self.P("  Connection retry {}/{} of connect session {}, total retries {}:".format(
          nr_retry, self.cfg_max_retries, self.nr_connection_issues // self.cfg_max_retries, self.nr_connection_issues,
        ))
        self._capture = self._open_stream()

        if self._capture is None or not self._capture.isOpened():
          self.P("    Capture failed.", color='r')
          continue

        has_frame, frame = self._capture.read()
        if not has_frame:
          self.P("    Capture read failed.", color='r')
          continue

        if self.cfg_nr_skip_frames:
          self._capture.set(cv2.CAP_PROP_POS_FRAMES, self.cfg_nr_skip_frames)

        _fps = 20
        try:
          _fps = int(self._capture.get(cv2.CAP_PROP_FPS))
        except:
          pass

        fps = np.clip(_fps, 15, 40)
        buff_size = self._capture.get(cv2.CAP_PROP_BUFFERSIZE)
        height = int(self._capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        width = int(self._capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        frame_count = int(self._capture.get(cv2.CAP_PROP_FRAME_COUNT))
        frame_count = max(0, frame_count)
        self.P("""Video stream configuration:
          AMD Simulate/FPS: {}/{}
          FPS={} (reported: {})
          Shape (HxW): {}
          Frame count: {}
          Buff size:   {} ({})""".format(
          self.cfg_simulate_amd, self.cfg_amd_target_dps,
          fps, _fps, (height, width), frame_count, buff_size, buff_size.__class__.__name__,
        ))

        self._metadata.fps = fps
        self._metadata.frame_h = height
        self._metadata.frame_w = width
        self._metadata.buffersize = buff_size

        # universal video stream code
        self._maybe_configure_frame_crop_resize(
          height=height,
          width=width,
        )
        # end universal video stream code

        self._metadata.frame_count = frame_count
        self.has_connection = True
      except:
        str_e = traceback.format_exc()
        self.P('`_maybe_reconnect` exception: {}'.format(str_e), color='r')
      # end try-except

      if self.has_connection:
        self.P("Connection seems to be established.")
        break
    # endwhile

    # random sleep time to de-sync broken streams
    sleep_time = np.random.randint(self.cfg_sleep_time_on_error, self.cfg_sleep_time_on_error * 3)

    if self.has_connection:
      self.reset_received_first_data()
      msg = "Video DCT '{}' successfully connected. Overall {} reconnects.".format(
        self.cfg_name,
        self.nr_connection_issues,
      )
      notification_type = ct.STATUS_TYPE.STATUS_NORMAL
      color = 'g'
    else:
      msg = "Abnormal functioning of Video DCT '{}' failed after {} retries, overall {} reconnects. Sleeping DCT for {:.1f}s".format(
        self.cfg_name, self.cfg_max_retries,
        self.nr_connection_issues,
        sleep_time,
      )
      notification_type = ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING
      color = 'e'
    # endif
    self.P(msg, color=color)
    self._create_notification(
      notif=notification_type,
      msg=msg,
      info=str_e,
      video_stream_info=self.deepcopy(self._metadata.__dict__),
      displayed=True,
    )

    if not self.has_connection:
      # execute sleep after notification delivery
      sleep(sleep_time)
    return

  def _get_stream_fps_max_thr(self):
    return self._metadata.fps * 0.6

  def _get_nr_grabs(self):
    if self.cfg_amd_target_dps > 0:
      n_grabs = int(np.ceil(self.cfg_cap_resolution / self.cfg_amd_target_dps))
    else:
      n_grabs = int(np.ceil(self._metadata.fps / self.cap_resolution)) + 1
    return n_grabs

  def _recalc_cap_resolution(self):
    # _heuristic_cap_resolution recalc for AMD (so far)
    WARM_UP_SEC = 60
    MAX_TIME = 2
    if self.cap_resolution >= self._get_stream_fps_max_thr():
      nr_streams = self.get_nr_parallel_captures() + 5
      read_time = self.get_timer('cv2_read')['MEAN']
      read_count = self.get_timer('cv2_read')['COUNT']
      if read_count > (self.cap_resolution * WARM_UP_SEC):
        total_read_time = read_time * self.cap_resolution * nr_streams
        # TODO: change below to more progressive & adaptive method
        if total_read_time >= MAX_TIME:
          configured_cap_res = self.get_cap_or_forced_resolution()
          self._heuristic_cap_resolution = round(MAX_TIME / (read_time * nr_streams), 1)
          self.P("Total cv2r time @{} load: {:.4f}s({:.4f}s/strm/itr) exceding {}s. Forcing cap {} @ {:.1f} dps, nr_grabs={} (stream fps: {} cap dps:{}/{})".format(
            nr_streams, total_read_time, read_time, MAX_TIME,
            self.cfg_name, self._heuristic_cap_resolution,
            self._get_nr_grabs(),
            self._metadata.fps,
            self.cap_resolution,
            configured_cap_res,
          ),
            color='m')
    return

  def _capture_read(self):
    has_frame, frame = False, None

    # COMMENT: a naive approach will not work in real-life multi-stream scenarios
    # has_frame, frame = self._capture.read()
    # END COMMENT: do not remove this

    if (self.cap_resolution >= self._get_stream_fps_max_thr() and self.cfg_amd_target_dps == 0) or self.is_intel():
      self.start_timer('cv2_read')  # DO NOT REMOVE - needs to "wrap" custom dps reads for _recalc_cap_resolution
      self.start_timer('cv2_read_dps{}'.format(self.cap_resolution))
      has_frame, frame = self._capture.read()
      self.end_timer('cv2_read_dps{}'.format(self.cap_resolution))
      self.end_timer('cv2_read')
    else:
      self.start_timer('cv2_grab_retrv_dps{}'.format(self.cap_resolution))
      nr_grabs = self._get_nr_grabs()
      self.start_timer('cv2_grab_x{}'.format(nr_grabs))
      for _ in range(nr_grabs):
        self.start_timer('cv2_grab')
        try:
          _ = self._capture.grab()
        except:
          break
        self.end_timer('cv2_grab')
      # endfor all grabs
      self.end_timer('cv2_grab_x{}'.format(nr_grabs))
      self.start_timer('cv2_retrieve')
      has_frame, frame = self._capture.retrieve()
      self.end_timer('cv2_retrieve')
      self.end_timer('cv2_grab_retrv_dps{}'.format(self.cap_resolution))
    if not self.is_intel() and self.cfg_amd_target_dps == 0:
      self._recalc_cap_resolution()
    return has_frame, frame

  def _run_data_aquisition_step(self):
    if self.has_connection and self._capture.isOpened():
      has_frame, frame = self._capture_read()
      if has_frame:

        # universal video stream code
        frame = self._maybe_resize_crop(frame)
        # end universal video stream code

        frame = np.ascontiguousarray(frame[:, :, ::-1])  # transform to rgb & contiguous

        # check for HW issues on some nvr/dvr
        if ((self.cfg_configured_h > 0 and self.cfg_configured_h != frame.shape[0])
            or
                (self.cfg_configured_w > 0 and self.cfg_configured_w != frame.shape[1])):
          if not self.__configured_size_error:
            self.__configured_size_error = True
            msg = "ERROR: Video stream configured at {}x{} is decoded at {}x{}".format(
              self.cfg_configured_h, self.cfg_configured_w,
              frame.shape[0], frame.shape[1],
            )
            self._create_notification(
              notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
              stream_name=self.cfg_name,
              msg=msg,
            )
          # endif requires exception notification
        # endif maybe we know specific HW for the video stream

        # metadata
        self._crt_frame += 1
        self._metadata.frame_current = self._crt_frame

        self._add_inputs(
          [
            self._new_input(img=frame, metadata=self._metadata.__dict__.copy()),
          ]
        )
      else:
        self.has_finished_acquisition = True
    else:
      self.P('Capture seems to be closed. Setting `has_connection=False`', color='r')
      self.has_connection = False
    return



import tensorflow as tf
from tensorflow.keras import initializers, activations
from tensorflow.python.keras.applications import imagenet_utils

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        'distribution': 'truncated_normal'
    }
}

class SmartSeparableConv2D(tf.keras.layers.Layer):
  def __init__(self,
               filters,
               dw_kernel_size,
               dw_strides,
               activation,
               expand_ratio=0,
               se_ratio=0.25,
               kernel_initializer=CONV_KERNEL_INITIALIZER,
               use_pointwise_batch_norm=True,
               **kwargs):
    super(SmartSeparableConv2D, self).__init__(**kwargs)    
    assert se_ratio > 0 and se_ratio <= 1
    
    if kernel_initializer != CONV_KERNEL_INITIALIZER:
      kernel_initializer = initializers.get(kernel_initializer)
      
    self.filters = filters
    self.dw_kernel_size = dw_kernel_size
    self.dw_strides = dw_strides
    self.expand_ratio = expand_ratio
    self.se_ratio = se_ratio
    self.activation = activations.get(activation)
    self.kernel_initializer = kernel_initializer
    self.use_pointwise_batch_norm = use_pointwise_batch_norm
    
  def _correct_pad(self, input_shape, kernel_size):
    """Returns a tuple for zero-padding for 2D convolution with downsampling.
      Arguments:
      inputs: Input tensor.
      kernel_size: An integer or tuple/list of 2 integers.
    
      Returns:
        A tuple.
    """
    input_size = input_shape[1:3]
    if isinstance(kernel_size, int):
      kernel_size = (kernel_size, kernel_size)
    if input_size[0] is None:
      adjust = (1, 1)
    else:
      adjust = (1 - input_size[0] % 2, 1 - input_size[1] % 2)
    correct = (kernel_size[0] // 2, kernel_size[1] // 2)
    return ((correct[0] - adjust[0], correct[0]),
            (correct[1] - adjust[1], correct[1]))
  
  def build(self, input_shape):
    filters_in  = int(input_shape[-1])    
    filters_se  = max(1, int(filters_in * self.se_ratio))
    
    #expand component layers
    if self.expand_ratio > 0:
      filters_in = filters_in * self.expand_ratio
      self.exp_conv   = tf.keras.layers.Conv2D(filters=filters_in,
                                               kernel_size=1,
                                               padding='same',
                                               use_bias=False,
                                               kernel_initializer=self.kernel_initializer,
                                               name=self.name + 'expand_conv')
      self.exp_bn     = tf.keras.layers.BatchNormalization(name=self.name + 'expand_bn')
      self.exp_act    = tf.keras.layers.Activation(self.activation, name=self.name + 'expand_act')    
    #endif
    
    #depthwise component layers
    dw_padding = 'same'
    if self.dw_strides == 2:
      self.dwconv_pad = tf.keras.layers.ZeroPadding2D(padding=self._correct_pad(input_shape, self.dw_kernel_size),
                                                      name=self.name + 'dwconv_pad')
      dw_padding = 'valid'
    #endif
    self.dwconv     = tf.keras.layers.DepthwiseConv2D(kernel_size=self.dw_kernel_size,
                                                      strides=self.dw_strides,
                                                      padding=dw_padding,
                                                      use_bias=False,
                                                      depthwise_initializer=self.kernel_initializer,
                                                      name=self.name + 'dwconv')
    self.dwconv_bn  = tf.keras.layers.BatchNormalization(name=self.name + 'bn')
    self.dwconv_act = tf.keras.layers.Activation(self.activation, name=self.name + 'act')
    
    #squeeze-excite component
    self.se_gap     = tf.keras.layers.GlobalAveragePooling2D(name=self.name + 'se_squeeze')
    self.se_res     = tf.keras.layers.Reshape(target_shape=(1, 1, filters_in), 
                                              name=self.name + 'se_reshape')
    self.se_reduce  = tf.keras.layers.Conv2D(filters=filters_se,
                                             kernel_size=1,
                                             padding='same',
                                             kernel_initializer=self.kernel_initializer,
                                             activation=self.activation,
                                             name=self.name + 'se_reduce')
    self.se_expand  = tf.keras.layers.Conv2D(filters=filters_in,
                                             kernel_size=1,
                                             padding='same',
                                             kernel_initializer=self.kernel_initializer,
                                             activation='sigmoid',
                                             name=self.name + 'se_expand')
    self.se_excite  = tf.keras.layers.Multiply(name=self.name + 'se_excite')
    
    #pointwise component
    self.pw_conv    = tf.keras.layers.Conv2D(filters=self.filters,
                                             kernel_size=1,
                                             padding='same',
                                             use_bias=False,
                                             kernel_initializer=self.kernel_initializer,
                                             name=self.name + 'project_conv')
    if self.use_pointwise_batch_norm:
      self.pw_bn      = tf.keras.layers.BatchNormalization(name=self.name + 'project_bn')
    
    super(SmartSeparableConv2D, self).build(input_shape)
    
  def call(self, inputs):
    tf_x  = inputs
    if self.expand_ratio > 0:
      tf_x = self.exp_conv(tf_x)
      tf_x = self.exp_bn(tf_x)
      tf_x = self.exp_act(tf_x)
    if self.dw_strides > 1:
      tf_x = self.dwconv_pad(tf_x)
    tf_x  = self.dwconv(tf_x)
    tf_x  = self.dwconv_bn(tf_x)
    tf_x  = self.dwconv_act(tf_x)
    tf_se = self.se_gap(tf_x)
    tf_se = self.se_res(tf_se)
    tf_se = self.se_reduce(tf_se)
    tf_se = self.se_expand(tf_se)
    tf_x  = self.se_excite([tf_x, tf_se])
    tf_x  = self.pw_conv(tf_x)
    if self.use_pointwise_batch_norm:
      tf_x  = self.pw_bn(tf_x)
    tf_out = tf_x
    return tf_out
  
  def get_config(self):
    config = {
        'activation'              : self.activation,
        'kernel_initializer'      : self.kernel_initializer,
        'filters'                 : self.filters,
        'dw_kernel_size'          : self.dw_kernel_size,
        'dw_strides'              : self.dw_strides,
        'expand_ratio'            : self.expand_ratio,
        'se_ratio'                : self.se_ratio,
        'use_pointwise_batch_norm': self.use_pointwise_batch_norm
        }
    base_config = super(SmartSeparableConv2D, self).get_config()
    cfg =  dict(list(base_config.items()) + list(config.items()))
    return cfg   

  def model(self):
    x = tf.keras.layers.Input(shape=(100, 200, 3), name='inp')
    model = tf.keras.models.Model(inputs=[x], outputs=self.call(x))
    return model
    
if __name__ == '__main__':
  tf_inp = tf.keras.layers.Input(shape=(300, 300, 3), name='inp')
  tf_x = tf_inp
  tf_x = SmartSeparableConv2D(filters=32, 
                              dw_kernel_size=3, 
                              dw_strides=2,
                              expand_ratio=6,
                              activation='relu',
                              name='ssc2d')
  # model = tf.keras.models.Model(tf_inp, tf_x)
  # model.summary()
  
  tf_x.build(input_shape=(None, 100, 200, 3))
  tf_x.model().summary()
  
  
  
  
  
  
  


#global dependencies
import os
import numpy as np
import abc

from time import sleep, time
from threading import Thread
from collections import deque
from datetime import datetime

#local dependencies
from naeural_core import constants as ct

from naeural_core.data_structures import MetadataObject
from naeural_core.data.base import BaseDataCapture
from naeural_core.data.mixins_base import _AquisitionOnIntervalsMixin, _DCTUtilsMixin



RUN_ON_THREAD = True
REGISTER_TIMERS = True

_CONFIG = {
  **BaseDataCapture.CONFIG,
  
  'IS_THREAD' : True,
  
  'CAP_RESOLUTION'        : 1,
  
  'FORCE_CAP_RESOLUTION'  : -1, # if -1, this parameter is ignored
  
  'RECONNECTABLE'         : 'YES',
  
  
  'SLEEP_TIME_ON_ERROR'   : 10,
  'MAX_RETRIES'           : 10,
  'MAX_DEQUE_LEN'         : 128, # 1 if 'LIVE_FEED': true
  'STREAM_WINDOW'         : 1, # if 'LIVE_FEED': true, this has no effect
  
  'LIVE_FEED'             : True, # default true
  
  'NR_SKIP_FRAMES'        : None, 
  
  ## SERIALIZATION & PERSISTENCE
  'LOAD_PREVIOUS_SERIALIZATION'   : False,
  'SERIALIZATION_SIGNATURE'       : None,
  ## END SERIALIZATION & PERSISTENCE
  
  
  'VALIDATION_RULES': {
    **BaseDataCapture.CONFIG['VALIDATION_RULES'],

    'CAP_RESOLUTION': {
      'TYPE': 'float',
      'MIN_VAL': 0.0003, 
      'MAX_VAL': 250,
      'DESCRIPTION': 'Sets the frequency of the DCT plugin where the minimum is once every hour and the maximum is 250 times per second.'
    },
    
    'RECONNECTABLE': {
      'TYPE': ['str', 'bool'],
    
      'DESCRIPTION': """
      Sets behaviour for connection/data stopping events: 
      - "YES"/true will force reconnecting when connection is lost, 
      - "NO"/false will finish the data capture and the included underlying pipeline with all its plugins if the data has finished or the connection is lost,
      - "KEEPALIVE" when the connection is lost or data has finished, the pipeline will enter a zombie state where no data is produced yet the business 
      plugins are still runnning and the pipeline can only be closed with an external command
      """
    }
  }
}

class DataCaptureThread(BaseDataCapture, 
                        _AquisitionOnIntervalsMixin,
                        _DCTUtilsMixin,
                        ):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self._phase = None
    self._deque = None
    self._thread = None
    self._stop = False
    self._register_timers = REGISTER_TIMERS
    self._DPS = -1
    
    self.__last_get_data_capture_timestamp = 0
    self.__get_data_capture_laps = deque(maxlen=1000)

    self.thread_stopped = False
    self.has_connection = False
    self.has_finished_acquisition = False
    self.__lap_times = deque(maxlen=100)
    self.__process_times = deque(maxlen=100)
    self.__sleep_times = deque(maxlen=100)

    self.__has_received_data = False
    self._start_time = time()
    self._timers_section = None
    self.__last_add_input_timestamp = time()

    self._metadata = MetadataObject(
      current_interval=None,
      payload_context=None,
      use_local_comms_only=False,
    )

    super(DataCaptureThread, self).__init__(**kwargs)
    return
  
  
  @property
  def _cache_folder(self):
    return ct.CACHE_DATA_CAPTURE
    
  @property
  def plugin_id(self):
    str_type = self.cfg_type
    str_name = self.cfg_name
    str_id = "{}__{}".format(str_type, str_name)
    return self.sanitize_name(str_id)

    
  def sleep(self, s):
    sleep(s)
    return
  

  def startup(self):
    super().startup()

    self._timers_section = 'STREAM_' + self.cfg_name
    maxlen = 1 if self.cfg_live_feed else max(self.cfg_max_deque_len, self.cfg_stream_window)
    self._deque = deque(maxlen=maxlen)

    crt_log = "Starting {} -> {} with following options:".format(self._signature, self.cfg_name)
    crt_log += "\n  Queue size:      {}".format(maxlen)
    crt_log += "\n  Capture res:     {} d/s".format(self.cap_resolution)
    crt_log += "\n  Is reconn.:      {}".format(self.is_reconnectable)
    crt_log += "\n  Skip data:       {}".format(self.cfg_nr_skip_frames)
    for key, val in self.config.items():
      if key not in [ct.PLUGINS, ct.CAP_RESOLUTION, ct.NR_SKIP_FRAMES, "VALIDATION_RULES"] and val is not None:
        crt_log += "\n  {:<15} {}".format(key + ':', val)

    crt_warn = ''
    if not self.cfg_live_feed and not self.is_reconnectable:
      crt_warn += '\n[Warning] Be aware that stream will be closed when all the data in buffer is consumed!'

    if self.cfg_live_feed and self.cfg_stream_window > 1:
      crt_warn += '\n[Warning] Be aware that using a stream window in live feed does not affect anything'

    self.P(crt_log, color='y') 
    if len(crt_warn) > 0:
      self.P(crt_warn, color='r')
    return
  

  def _populate_stream_metadata_from_config(self):
    """
    This protected method populates the `_stream_metadata` object with basic information
    at initiation time. Part of the information is also added in capture getter and should 
    be moved.
    
    Should be populated in derived classes.

    Returns
    -------
    None.

    """
    super()._populate_stream_metadata_from_config()    
    self._stream_metadata.url = self.cfg_url
    return
  

  @property
  def is_reconnectable(self):
    return (
      (isinstance(self.cfg_reconnectable, str) and  self.cfg_reconnectable.upper() in ['YES','Y', 'T', 'TRUE']) or 
      (isinstance(self.cfg_reconnectable, bool) and self.cfg_reconnectable)
    )
    
  
  @property
  def is_keepalive(self):
    return isinstance(self.cfg_reconnectable, str) and  self.cfg_reconnectable.upper() in ['KEEPALIVE', 'KEEP_ALIVE']
  
    
  def get_cap_or_forced_resolution(self):
    cap_resolution = self.cfg_cap_resolution
    force_cap_resolution = self.cfg_force_cap_resolution
    if force_cap_resolution > 0:
      cap_resolution = force_cap_resolution
    return cap_resolution

  
  @property
  def has_cap_resolution_config(self):
    cap = self.config.get(ct.CAP_RESOLUTION, None)
    is_valid = cap is not None and isinstance(cap, (int, float)) and cap > 0
    return is_valid
  
  
  @property
  def heuristic_cap(self):
    if hasattr(self, '_heuristic_cap_resolution'):
      if self._heuristic_cap_resolution is not None and self._heuristic_cap_resolution > 0:
        return round(self._heuristic_cap_resolution, 2)
    return None


  @property
  def cap_resolution(self):
    """
    Use this property-function if you need to use the heuristics / auto-calculation of DPS
    otherwise you can use directly get_proposed_cap_resolution

    Returns
    -------
    TYPE: float
      required number of stream reads per second.

    """
    heuristic_cap = self.heuristic_cap
    default_cap = self.get_cap_or_forced_resolution()  
    if heuristic_cap is not None:
      return heuristic_cap
    else:
      return default_cap


  @property
  def generate_resolution(self):
    """
    Data generation resolution is the number of times the data is generated per second for
    downstream by the get_data_capture method. This is different from the cap_resolution
    """
    if len(self.__get_data_capture_laps) > 0:
      return round(1 / np.mean(self.__get_data_capture_laps), 2)
    else:
      return -1
    


  @property
  def can_delete(self):
    # return True if stream can reconnect or someone else needs to destroy it
    return self.thread_stopped and not (self.is_reconnectable or self.is_keepalive)


  @property
  def is_thread(self):
    return True


  @property
  def actual_dps(self):
    return self._DPS

  
  @property
  def time_since_last_input(self):
    return time() - self.__last_add_input_timestamp
  

  @property
  def received_first_data(self):
    return self.__has_received_data
  

  def start_timer(self, tmr_id):
    if not self._register_timers:
      return

    self.log.start_timer(sname=tmr_id, section=self._timers_section)
    return

  def get_timer(self, tmr_id):
    return self.log.get_timer(skey=tmr_id, section=self._timers_section)

  def end_timer(self, tmr_id, skip_first_timing=True, periodic=False):
    if not self._register_timers:
      return -1

    return self.log.end_timer(sname=tmr_id, section=self._timers_section, skip_first_timing=skip_first_timing, periodic=periodic)
  
  
  def reset_received_first_data(self):
    self.__has_received_data = False
    return
  
  def mark_receved_first_data(self):
    self.__has_received_data = True
    return
  

  def _new_input(self, img=None, struct_data=None, metadata=None, init_data=None):
    """
    Utilitary for creating a sub-stream input

    *** Very important observation: ***
      `img` and `struct_data` are mutual exclusive (i.e. a sub-stream can collect either a frame, or a chunk of structured data)

    Parameters:
    ----------
    img: np.ndarray, optional
      The image collected by the sub-stream
      The default is None.

    struct_data: object, optional
      The chunk of structured data collected by the sub-stream
      The default is None

    metadata: dict, optional
      Metadata of the sub-stream
      The default is None ({})

    init_data: object, optional
      Some initial data collected by the sub-stream (training data or anchor images for example)
      The default is None

    Returns:
    --------
    An input in the DecentrAI format - a dictionary as following:
      {
        "IMG"         : $(the received value in `img` parameter),
        "STRUCT_DATA" : $(the received value in `struct_data` parameter),
        "METADATA"    : $(the received value in `metadata` parameter),
        "INIT_DATA"   : $(the received value in `init_data` parameter),
        "TYPE"        : "IMG" or "STRUCT_DATA" ### specifies whether the field "IMG" or the field "STRUCT_DATA" has value
      }

    """

    if img is not None:
      assert struct_data is None
      _type = 'IMG'

    if struct_data is not None or init_data is not None:
      assert img is None
      _type = 'STRUCT_DATA'

    _metadata = metadata
    if _metadata is None:
      if hasattr(self, "_metadata") and isinstance(self._metadata, MetadataObject):
        _metadata = self._metadata.__dict__.copy()
      else:
        _metadata = {}
    _metadata = {
      **_metadata,
      # **self.cfg_stream_config_metadata,
    }
    return {
      "IMG" : img,
      "STRUCT_DATA" : struct_data,
      "METADATA" : _metadata,
      "INIT_DATA" : init_data,
      "TYPE" : _type
    }

  
  def __on_init(self):
    self._on_init()
    return
    
  def on_init(self):
    """
    Called before the normal connection is established. This method should be used to initialize
    variables.
    """
    return
  
  def _on_init(self):
    self.on_init()
    return  


  def _init(self):
    self._maybe_reconnect()
    return
      
  
  def _maybe_reconnect(self):
    if self.has_connection:
      return    
    if self.connect():
      self.has_connection = True  
    else:
      self.has_connection = False
    return self.has_connection
  
  @abc.abstractmethod
  def connect(self):
    raise NotImplementedError()
  
  @abc.abstractmethod
  def _release(self):
    raise NotImplementedError()


  def _run_data_aquisition_step(self):
    """
    Runs a data acquisition step. This method is called in the main loop of the DCT.
    It is responsible for collecting data from the source and adding it to the deque.
    
    
    The `data_step` method is the main method that should be implemented in the derived classes. 
    """    
    if hasattr(self, 'data_step'):
      self.data_step()
    elif hasattr(self, 'run_data_aquisition_step'):
      self.run_data_aquisition_step()
    else:
      raise NotImplementedError("Method `data_step` must be implemented in the derived class")
    return
  
  def __run_data_acquisition_step(self):
    self._run_data_aquisition_step()
    return

  def _handle_commands(self):
    # TODO: rethink command structure
    commands = self.config.get('COMMANDS', [])
    for cmd in commands:
      # check each command and execute specific setup
      known = False
      if cmd.get('FINISH_ACQUISITION', False):
        # basically a DCT loop exit
        self.has_finished_acquisition = True
        received_cmd = "FINISH_ACQUISITION"
        known = True
      else:
        received_cmd = cmd
        known = False
      # endif any known command check 
      msg = "Received {} '{}' command from INITIATOR_ID:'{}' with SESSION_ID:'{}'".format(
        'known' if known else 'unknown',
        received_cmd,
        self.config.get('INITIATOR_ID'),
        self.config.get('SESSION_ID'),
        )
      self.P(msg, color='r' if not known else None)
    #endfor each command
    return

  def _add_struct_data_input(self, obs):
    """
    This method is used to add a structured data input to the stream.

    Parameters
    ----------
    obs : any
        the structured data to be added to the stream.
    """
    self._add_inputs(
      [
       self._new_input(img=None, struct_data=obs, metadata=self._metadata.__dict__.copy(), init_data=None),
      ]
    )   
    
    
  def _add_struct_data_inputs(self, obs_list):
    """
    This method is used to add a list of structured data inputs to the stream.

    Parameters
    ----------
    obs_list : list
        list of any structured data to be added to the stream.
    """
    inputs = []
    for obs in obs_list:
      inputs.append(self._new_input(img=None, struct_data=obs, metadata=self._metadata.__dict__.copy(), init_data=None))
    self._add_inputs(inputs)
    return
  
    
  def _add_img_input(self, img):
    self._add_inputs(
      [
       self._new_input(img=img, struct_data=None, metadata=self._metadata.__dict__.copy(), init_data=None),
      ]
    )    
    

  def _add_inputs(self, inputs):
    """
    Threaded routine for adding stream data in the deque.

    Parameters
    ----------
    inputs : list[dict]
      A list with input for each sub-stream in the current stream.
      Each input (dictionary) is a structure returned by `self._new_input(...)`
    """

    assert isinstance(inputs, list)

    if not self.cfg_live_feed:
      #if in `buffer` mode and the buffer is full, wait the buffer to release a position and then add a new item
      tm_add_inputs_start = time()
      while True:
        # break loop if stop or config change
        if self._stop or self._loop_paused:
          self.P("Skipping {} frames! STOP:{}  LOOP_PAUSED:{}".format(len(inputs), self._stop, self._loop_paused), color='r')
          break
        # very important: update if maxlen==1 although this is already taken care by self.cfg_live_feed
        elif ((len(self._deque) == self._deque.maxlen) and (self._deque.maxlen > 1)): 
          self.sleep(0.05)
          # # now check if too much time wasted          
          # TODO(AID): add a flag that enables this
          # if (time() - tm_add_inputs_start) > 2:
          #   # maybe too much time wasted here so break it
          #   break
        else:
          break
        #endif
      #endwhile
    #endif
    self._deque.append(inputs)
    self.__last_add_input_timestamp = time()
    return
  
  # def _clear_inputs(self):
  #   """
  #   Use this method if you want to invalidate all data in the deque.
  #   This method should be used only in very specific situations, such as in a URL change,
  #   where the business plugin 
  #   """
  #   self._deque.clear()
    
  
  def _reset_state(self):
    self.has_connection = False
    self.has_finished_acquisition = False
    self.thread_stopped = False
    return

  def _can_run_data_aquisition_step(self):
    current_interval = self.acquisition_on_intervals_current_interval()
    if isinstance(current_interval, str) and len(current_interval) == 0:
      return False, current_interval
    return True, current_interval

  def _run_thread(self):
    try:
      self.__on_init()
      self._init() 
      lap_timestamp = None
      while True:

        while self._loop_paused:
          sleep(0.001)
          self._capture_thread_waiting = True
        
        # first trigger on-config and set LAST_UPDATE_TIME 
        self.maybe_trigger_on_config()
          
        self._capture_thread_waiting = False

        self._capture_loop_in_exec = True
        start_processing_time = time()
        self.start_timer('thread_loop')
        #  _maybe_reconnect must be defined by each individual DCT
        self._maybe_reconnect()

        # first run commands in any (this will also take care of LAST_UPDATE_TIME)
        self.__maybe_trigger_pipeline_command()
        
        # now run normal data acquisition                
        can_run, current_interval = self._can_run_data_aquisition_step()
        if can_run:
          self._metadata.current_interval = current_interval
          self.start_timer('run_data_acquisition_step')
          #  _run_data_aquisition_step must be defined by each individual DCT
          self.__run_data_acquisition_step()
          self.end_timer('run_data_acquisition_step')
        #endif
        self._capture_loop_in_exec = False

        if lap_timestamp is None:
          lap_timestamp = time()
        else:
          lap_time = time() - lap_timestamp
          self.__lap_times.append(lap_time)
          avg_time = np.mean(self.__lap_times)
          self._DPS = 1 / avg_time
          lap_timestamp = time()
        
        if self._stop:
          #stop command received from outside. stop imediatly
          self.P('`stop` command received. Exiting from `{}.run_thread`'.format(
            self.__class__.__name__))
          break
        elif self.has_finished_acquisition:
          if not self.is_reconnectable or self.is_keepalive:            
            # no more data to collect, no reconnection needed, then exit
            # at this point we also break from keep-alive streams that need to stay 
            # zombie (even with no data) until some cleanup happends, so while "NO"
            # streams will be collected for deletion the keep-alive ones will not be
            # collected
            self.P('`_run_data_aquisition_step` finished. Exiting from `run_thread`')
            break
          else:
            #reset states and allow main loop to restart the process
            self._reset_state()
        #endif
        ###
        self._handle_commands()
        ###
        processing_time = time() - start_processing_time
        self.__process_times.append(processing_time)
        # give other threads some time or predefined time
        self.start_timer('thread_sleep')
        # next we calc sleep based on cap_resolution property that choses
        # between proposed and heuristical resolution (if available)
        sleep_time = max(1 / self.cap_resolution - processing_time, 0.00001)
        sleep(sleep_time)
        elapsed_sleep = self.end_timer('thread_sleep')
        self.__sleep_times.append(elapsed_sleep)
        self.end_timer('thread_loop')
      #endwhile
      
      # wait for deque to be fully read (LIVE_FEED=false) if thread was not 
      # force-stopped, so this way we ensure the downstream tasks are aware 
      # and receive the last observation
      if len(self._deque) > 0 and not self._stop:
        self.P("thread waiting for manager to consume the queue...")
      while len(self._deque) > 0 and not self._stop:
        sleep(1)
      
      # THIS SHOULD BE DEFINED IN SUBCLASS
      self._release()
      # END
      
      self.P('DCT `run_thread` finished')
      self.thread_stopped = True
    except Exception as exc:
      exc_info = self.trace_info()
      msg = "EXCEPTION in capture's _run_thread: {}".format(exc)
      self.P(msg, color='error')
      self.P("  >> info:\n {}".format(exc_info), color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info=exc_info,
        autocomplete_info=True
      )
    #end try-except whole DCT loop
    # now that loop exists we can safely perform additional cleanup
    return
  
  
  def get_runstats(self, last_n=1):
    np_laps = np.array(self.__lap_times).round(3)
    np_proc = np.array(self.__process_times).round(3)
    np_sleep = np.array(self.__sleep_times).round(3)    
    status = ''
    np.set_printoptions(precision=3, floatmode='fixed')
    if len(np_laps) > 1 and len(np_proc) > 1 and len(np_sleep) > 1:
      status = "{:.03f} {} = {:.03f} {} + {:.03f} {}".format(
        np_laps.mean(), np_laps[-last_n:],
        np_proc.mean(), np_proc[-last_n:],
        np_sleep.mean(), np_sleep[-last_n:],
      )
    return status

  def get_data_capture(self):
    if self.has_data and not self.received_first_data:
      self.P("Data received from source after connect.", color='g')
      self.mark_receved_first_data()

    if self.cfg_live_feed:
      all_inputs = self._deque.pop()
    else:
      all_inputs = []
      # TODO: review below section
      for _ in range(self.cfg_stream_window):
        if len(self._deque) == 0:
          break
        all_inputs += self._deque.popleft()
      #endfor
    #endif

    metadata = self._stream_metadata
    metadata.actual_dps = round(self.actual_dps, 2)
    
    metadata.cap_resolution = self.cap_resolution
    metadata.cap_max_queue_len = self._deque.maxlen
    metadata.cap_queue_len = len(self._deque)
    metadata.cap_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
    metadata.cap_elapsed_time = round(time() - self._start_time, 2)
    metadata.cap_signature = self.__class__.__name__
    
    metadata.live_feed = self.cfg_live_feed
    metadata.reconnectable = self.cfg_reconnectable
    metadata = metadata.__dict__

    data_capture = self.data_template(all_inputs, metadata)
    
    if self.__last_get_data_capture_timestamp > 0:
      elapsed = self.time() - self.__last_get_data_capture_timestamp
      self.__get_data_capture_laps.append(elapsed)
    #endif
    self.__last_get_data_capture_timestamp = self.time()
    return data_capture

  @property
  def has_data(self):
    has_data = self._deque and len(self._deque) > 0
    return has_data
  
  def start(self):
    if RUN_ON_THREAD:
      self._thread = Thread(
        target=self._run_thread, 
        args=(), 
        name=ct.THREADS_PREFIX + 'data_' + self.cfg_name,
        daemon=True,      
      )
      self._thread.daemon = True
      self._thread.start()
    else:
      self.P("Warning! Data Capture Thread does not run on thread", color='e')
      self._run_thread()
    return
  
  def stop(self, join_time=10):
    self._stop = True
    if join_time:
      self._thread.join(join_time)
    self._deque.clear()
    self._reset_state()
    return  
    
    
  def _download(self, url, file_name=None, progress=False, verbose=False, notify_download=None, unzip=False):
    if file_name is None:
      file_name = self.log.now_str()
      
    path = self.log.check_folder_data(ct.FOLDER_DOWNLOADS)
    fn = os.path.join(ct.FOLDER_DOWNLOADS, file_name)
    self.log.maybe_download(
      url=url,
      fn=fn,
      target=ct.FOLDER_DATA,
      force_download=True,
      print_progress=progress, #set False to avoid blocking threads
      verbose=verbose, #set False to avoid blocking threads
      publish_func=notify_download,
      unzip=unzip
      )
    fn_out = os.path.join(path, file_name)
    return fn_out
  
  
  def _on_pipeline_command(self, cmd_data, payload_context=None, **kwargs):
    """This method must be implemented in derived classes if they need to handle pipeline commands

    Parameters
    ----------
    cmd_data : dict / str
        the actual data of the command
        
    payload_context : dict, optional
        the payload context of the command. The default is None.
        
    **kwargs : optional arguments derived from the command
    
    """
    return
  
  
  def __on_pipeline_command(self, cmd_data):
    self.P("Running `_on_pipeline_command` ({}:{}) with command: {}...".format(
      ct.CONFIG_STREAM.LAST_UPDATE_TIME, self.config_data.get(ct.CONFIG_STREAM.LAST_UPDATE_TIME, None),
      str(cmd_data)[:30])
    )
    payload_context = None
    kwargs = {} # should be populated with data from the payload if any
    if isinstance(cmd_data, dict):
      payload_context = cmd_data.get('PAYLOAD_CONTEXT', None)
    self._metadata.payload_context = payload_context # set the payload context

    # this part handles pipeline commands that should be answered only on the local comms 
    use_local_comms_only = False
    if isinstance(cmd_data, dict):
      use_local_comms_only = cmd_data.get('USE_LOCAL_COMMS_ONLY', False)
    self._metadata.use_local_comms_only = use_local_comms_only
    self._on_pipeline_command(
      cmd_data=cmd_data,
      payload_context=payload_context,
      **kwargs,
    )
    self._metadata.payload_context = None # reset the payload context
    self._metadata.use_local_comms_only = False # reset the use_local_comms_only
    return
  
  
  def __maybe_trigger_pipeline_command(self):
    triggered = False
    if self.cfg_pipeline_command is not None and len(self.cfg_pipeline_command)>0:
      # new command at pipeline level handled by the DCT!
      command = self.cfg_pipeline_command
      # make sure we save config before code
      self.archive_config_keys(keys=['PIPELINE_COMMAND'], defaults=[{}])
      # now run the code (on-config should have already been triggered)
      self.__on_pipeline_command(cmd_data=command)
      # copy, erase and save
      triggered = True
    #end has pipeline command
    return triggered
  
  
  def __save_config(self, keys):
    # writes the config for a set of particular keys on the local cache!
    save_config_fn = self.shmem[ct.CALLBACKS.PIPELINE_CONFIG_SAVER_CALLBACK]
    try:
      config = {k : self.config_data[k] for k in keys}
      save_config_fn(pipeline_name=self.cfg_name, config_data=config, skip_update_time=True)
      # the save will trigger a update config when the CaptureManager will get the new config from
      # the ConfigManager and will pass it as upstream_config to the plugin - basically will see that
      # we have the "last key" and the "current" keys modified
      # so we already update the upstream_config to avoid this
      # nevertheless we must `skip_update_time=True` in order to avoid the update time to be set by ConfigManager
      keys.append(ct.CONFIG_STREAM.LAST_UPDATE_TIME) # this triggers re-update of the config if not properly set
      for k in keys:
        self._upstream_config[k] = self.config_data[k]
    except Exception as exc:
      self.P("Error '{}' while saving keys {}".format(exc, keys))
      raise exc
    return
  
  
  def archive_config_keys(self, keys : list, defaults=None):
    """
    Method that allows resetting of a list of keys and saving the current value as `_LAST` keys

    Parameters
    ----------
    keys : list
      List of keys to be archived.

    defaults: list
      List of default values for all keys. Default is None

    Returns
    -------
    None.

    """
    if defaults is None:
      defaults = [None] * len(keys)
    assert len(defaults) == len(keys), "Default values must be provided for all keys"
    all_keys = keys.copy()
    # TODO: maybe ignore if LAST_PIPELINE_COMMAND contained 'IMG' key or if too large

    for i, key in enumerate(keys):
      archive_key = self.ct.CONFIG_STREAM.LAST_PIPELINE_COMMAND
      self.config_data[archive_key] = self.config_data[key]
      self.config_data[key] = defaults[i]
      all_keys.append(archive_key)
    #endfor
    self.save_config_keys(all_keys)
    return


  def save_config_keys(self, keys : list):
    """
    Method that allows saving the local config in local cache in order to update a
    specific set of given keys that might have been modified during runtime

    Parameters
    ----------
    keys : list
      List of keys to be saved.

    Returns
    -------
    None.

    """
    EXCEPTIONS = ['NAME', 'TYPE']
    result = False
    all_exist = all([k in self.config_data for k in keys])
    all_accepted = all([k not in EXCEPTIONS for k in keys])
    if all_exist and all_accepted:
      self.__save_config(keys)
      result = True
    return result


import json

from naeural_core.bc import DefaultBlockEngine
from naeural_core import Logger


if __name__ == '__main__':
  l = Logger('AENC', base_folder='.', app_folder='_local_cache')
  
  receiver = DefaultBlockEngine(
  log=l,
  name="jsreader",
  config= {
    "PEM_FILE"     : "jsreader.pem",
    "PASSWORD"     : None,      
    "PEM_LOCATION" : "data"
    },
  )
  
  str_output_payload = """
    {
      "EE_ENCRYPTED_DATA": true,
      "ENCRYPTED_DATA": "CjWJCMRV0GYWUOCjAeKEAOFbkp8pfFsWxE98MZHyoMpJx2lema7v+VwqcWNYy2b1ilwCcg==",
      "EE_SIGN": "MEUCIAufipoc4ltZUP9oo6sbhf4DXVmpOB09W5yaXTovDL4aAiEArOjuZ4TnscVtccHq2lwm7ch4x6Q7-C9o99yzkzrqMTQ=",
      "EE_SENDER": "0xai_AwwqvbL_Fw3y0MQzllx69JZSLYT3ybF9zanfrmcgAlEp",
      "EE_HASH": "b9ee0540599d52201dcc9b595b2bff7d35292882c4679ff5c89530a8b45539bf"
    }
  """
  

  str_incoming_payload = str_output_payload
  incoming_payload = json.loads(str_incoming_payload)

  is_verified = receiver.verify(
    dct_data=incoming_payload
  )  
  
  assert is_verified, "Signature verification failed"
  
  sender_address = incoming_payload.get("EE_SENDER")  
  incomind_enc_data = incoming_payload.get("ENCRYPTED_DATA")
  
  dec = receiver.decrypt(
    encrypted_data_b64=incomind_enc_data, 
    sender_address=sender_address,
    debug=True,
  )
  print(dec)


import posix_ipc
from mmap import mmap
import cv2
import numpy as np
import argparse
from xperimental.shm_config import SIZE_DEF, WIDTH_DEF, HEIGHT_DEF, CHANNELS_DEF, SHM_STREAMER_VAPOR, SHM_VAPOR_STREAMER

class Shm:
    def __init__(self, size, h, w, c, yuv=True):
        self._size = size
        self._h = h
        self._w = w
        self._c = c
        self._yuv = yuv

    def init_memory(self):
        self._memory_in = posix_ipc.SharedMemory(SHM_STREAMER_VAPOR, posix_ipc.O_RDWR, size=self._size)
        self._memory_out = posix_ipc.SharedMemory(SHM_VAPOR_STREAMER, posix_ipc.O_RDWR, size=self._size)

        self.mmap_in = mmap(self._memory_in.fd, self._memory_in.size)
        self.mmap_out = mmap(self._memory_out.fd, self._memory_out.size)

        self.buf_in = memoryview(self.mmap_in)
        self.buf_out = memoryview(self.mmap_out)

        if self._yuv:
            self.np_in = np.ndarray(shape=(int(self._h*1.5), self._w), dtype=np.uint8, buffer=self.buf_in)
            self.np_out = np.ndarray(shape=(int(self._h*1.5), self._w), dtype=np.uint8, buffer=self.buf_out)
        else:
            self.np_in = np.ndarray(shape = (self._h, self._w, self._c), dtype=np.uint8, buffer=self.buf_in)
            self.np_out = np.ndarray(shape = (self._h, self._w, self._c), dtype=np.uint8, buffer=self.buf_out)
    
    def np_read(self):
        np_rgb = self.np_in.copy()
        if self._yuv:
          np_rgb = cv2.cvtColor(np_rgb, cv2.COLOR_YUV2RGB_I420)
        return np_rgb

    def np_write(self, np_rgb):
        if self._yuv:
          np_rgb = cv2.cvtColor(np_rgb, cv2.COLOR_RGB2YUV_I420)          
        self.np_out[:] = np_rgb[:]

    def write_to_memory(self, mapfile, s):
        mapfile.seek(0)
        mapfile.write(s)

    def read_from_memory(self, mapfile):
        global frame_bytes_size
        mapfile.seek(0)
        return mapfile.read(frame_bytes_size) 

    def process_rgb_frame(self, frame):
        if type(frame) == bytes:
            img = np.frombuffer(frame, dtype=np.uint8)
            img = img.reshape((self._h,self._w,self._c))
            img = img[:,:,::-1]
        else:
            img = frame
        
        img = cv2.circle(img,(self._w//2, self._h//2), self._h//2, (0,0,255), 2)
        (text_w, text_h),_ = cv2.getTextSize(text="Pseudo AI processing" , fontFace=cv2.FONT_HERSHEY_TRIPLEX , fontScale=0.8 , thickness=2)
        left = self._w//2 - text_w//2
        bottom = self._h//2 + text_h//2
        img = cv2.putText(img, "Pseudo AI processing", (left, bottom), fontFace=cv2.FONT_HERSHEY_TRIPLEX, fontScale=0.8, thickness=2, color=(0,0,255))
        return img


def get_args(parser):
    size_arg = SIZE_DEF
    width_arg = WIDTH_DEF
    heigth_arg = HEIGHT_DEF
    channels_arg = CHANNELS_DEF
    args = parser.parse_args()
    if args.size != None:
        size_arg = args.size_yuv
    if args.w != None:
        width_arg = args.w
    if args.h != None:
        heigth_arg = args.h
    if args.c != None:
        channels_arg = args.c
    return size_arg, width_arg, heigth_arg, channels_arg

def add_args(parser):
    parser.add_argument('--size', help = 'Frame size in bytes')
    parser.add_argument('--w', help = 'Frame width in pixels')
    parser.add_argument('--h', help = 'Frame height in pixels')
    parser.add_argument('--c', help = 'Frame number of channels')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    add_args(parser)
    size, width, height, channels = get_args(parser)

    shm = Shm(size, height, width, channels, yuv=False)
    shm.init_memory()

    while True:
        frame = shm.np_read()
        frame = shm.process_rgb_frame(frame)
        shm.np_write(frame)
        #frame = shm.read_from_memory(shm.mmap_in)
        #frame = shm.process_rgb_frame(frame)
        #shm.write_to_memory(shm.mmap_out, frame)




import cv2 

if __name__ == '__main__':
  
  cap = cv2.VideoCapture('https://live.zootargoviste.ro:7001')
  done = False
  frames = 0
  while not done:
    has_frame, frame = cap.read()
    if has_frame:
      cv2.imshow("zoo", frame)
      frames += 1

    if cv2.waitKey(1) & 0xFF == ord('q'):
      done = True
        
  cap.release()
  cv2.destroyAllWindows()


# Resume all interrupted trainings in yolov5/ dir including DDP trainings
# Usage: $ python utils/aws/resume.py

import os
import sys
from pathlib import Path

import torch
import yaml

FILE = Path(__file__).resolve()
ROOT = FILE.parents[2]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

port = 0  # --master_port
path = Path('').resolve()
for last in path.rglob('*/**/last.pt'):
    ckpt = torch.load(last)
    if ckpt['optimizer'] is None:
        continue

    # Load opt.yaml
    with open(last.parent.parent / 'opt.yaml') as f:
        opt = yaml.safe_load(f)

    # Get device count
    d = opt['device'].split(',')  # devices
    nd = len(d)  # number of devices
    ddp = nd > 1 or (nd == 0 and torch.cuda.device_count() > 1)  # distributed data parallel

    if ddp:  # multi-GPU
        port += 1
        cmd = f'python -m torch.distributed.run --nproc_per_node {nd} --master_port {port} train.py --resume {last}'
    else:  # single-GPU
        cmd = f'python train.py --resume {last}'

    cmd += ' > /dev/null 2>&1 &'  # redirect output to dev/null and run in daemon thread
    print(cmd)
    os.system(cmd)



from naeural_core.core_logging import SBLogger
from naeural_core.local_libraries.nn.th.training.pipelines.autoencoder import AutoencoderTrainingPipeline

if __name__ == '__main__':
  log = SBLogger()

  CONFIG = {
    'MODEL_NAME' : 'MNIST_DAE',
    'PRELOAD_DATA' : True,
    'DEVICE_LOAD_DATA' : 'cpu',
    'DEVICE_TRAINING' : 'cpu',
    'BATCH_SIZE' : 32,
    'EPOCHS' : 20,

  }

  pipeline = AutoencoderTrainingPipeline(
    log=log,
    signature='autoencoder',
    config=CONFIG,
    path_to_dataset='./_local_cache/_data/MNIST'
  )

  pipeline.run()




from typing import Tuple
from enum import Enum
import torch as th
import torchvision as tv
import json

from naeural_core.xperimental.th_y8.graph_nms import y5_nms, y8_nms, y5_nms_topk, y8_nms_topk
from naeural_core.xperimental.th_y8.graph_nms import y8_nms_trt_efficient, y8_nms_onnx

from plugins.serving.architectures.y5.general import non_max_suppression as y5_nms_orig
# try:
#   from ultralytics.yolo.utils.ops import non_max_suppression as y8_nms_orig
# except:
#   print("EXCEPTION in loading original nms functions. Defaulting to graph ones.", flush=True)
#   y8_nms_orig = y8_nms

from naeural_core.xperimental.th_y8.graph_nms import xywh2xyxy

class BackendType(Enum):
  TORCH = 1
  TENSORRT = 2
  ONNX = 3
  OpenVINO = 4

def y8_nms_orig(
        prediction,
        # conf_thres:float=0.25,
        # iou_thres:float=0.45,
        # classes=None,
        # agnostic=False,
        # multi_label:bool=False,
        # max_det:int=300,
        # max_time_img:float=0.05,
        # max_nms:int=30000,
        # max_wh:int=7680,
):
  """
  Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

  Arguments:
      prediction (th.Tensor): A tensor of shape (batch_size, num_boxes, num_classes + 4 + num_masks)
          containing the predicted boxes, classes, and masks. The tensor should be in the format
          output by a model, such as YOLO.
      conf_thres (float): The confidence threshold below which boxes will be filtered out.
          Valid values are between 0.0 and 1.0.
      iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
          Valid values are between 0.0 and 1.0.
      classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
      agnostic (bool): If True, the model is agnostic to the number of classes, and all
          classes will be considered as one.
      multi_label (bool): If True, each box may have multiple labels.
      labels (List[List[Union[int, float, th.Tensor]]]): A list of lists, where each inner
          list contains the apriori labels for a given image. The list should be in the format
          output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
      max_det (int): The maximum number of boxes to keep after NMS.
      nc (int): (optional) The number of classes output by the model. Any indices after this will be considered masks.
      max_time_img (float): The maximum time (seconds) for processing one image.
      max_nms (int): The maximum number of boxes into tv.ops.nms().
      max_wh (int): The maximum box width and height in pixels

  Returns:
      (List[th.Tensor]): A list of length batch_size, where each element is a tensor of
          shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
          (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
  """

  # # Checks
  # assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
  # assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'

  # if isinstance(prediction, (list, tuple)):  # YOLOv8 model in validation model, output = (inference_out, loss_out)
  #     prediction = prediction[0]  # select only inference output

  # device = prediction.device
  # mps = 'mps' in device.type  # Apple MPS
  # if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
  #     prediction = prediction.cpu()
  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nm = 0 # prediction.shape[1] - nc - 4
  mi = 4 + nc  # mask start index
  xc = prediction[:, 4:mi].amax(1) > 0.25  # candidates  #conf_thres

  # Settings
  # min_wh = 2  # (pixels) minimum box width and height
  # time_limit = 0.5 + max_time_img * bs  # seconds to quit after
  # redundant = True  # require redundant detections
  # multi_label = False and (nc > 1) # multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
  # merge = False  # use merge-NMS

  # t = time.time()
  output = [th.zeros((0, 6 + nm), device=prediction.device)] * bs
  for xi in range(prediction.shape[0]):  # image index, image inference
    # Apply constraints
    # x[((x[:, 2:4] < min_wh) | (x[:, 2:4] > max_wh)).any(1), 4] = 0  # width-height
    x = prediction[xi]
    x = x.transpose(0, -1)[xc[xi]]  # confidence

    # Cat apriori labels if autolabelling
    # if labels and len(labels[xi]):
    #     lb = labels[xi]
    #     lbsz = len(lb)
    #     v = th.zeros((lbsz, nc + nm + 5), device=x.device)
    #     v[:, :4] = lb[:, 1:5]  # box
    #     v[th.range(lbsz), lb[:, 0].long() + 4] = 1.0  # cls
    #     x = th.cat((x, v), 0)

    # If none remain process next image
    if x.shape[0] > 0:

      # Detections matrix nx6 (xyxy, conf, cls)
      box, cls, mask = x.split((4, nc, nm), 1)
      box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)
      # if multi_label:
      #     i, j = (cls > conf_thres).nonzero(as_tuple=False).T
      #     x = th.cat((box[i], x[i, 4 + j, None], j[:, None].float(), mask[i]), 1)
      # else:  # best class only
      conf, j = cls.max(1, keepdim=True)
      x = th.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > 0.25]

      # # Filter by class
      # if classes is not None:
      #     x = x[(x[:, 5:6] == th.tensor(classes, device=x.device)).any(1)]

      # Apply finite constraint
      # if not th.isfinite(x).all():
      #     x = x[th.isfinite(x).all(1)]

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:30000]]  # sort by confidence and remove excess boxes #max_nms

        # Batched NMS
        c = x[:, 5:6] * 7680  # classes #max_wh
        boxes = x[:, :4] + c
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300]  # limit detections #max_det
        # if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
        #     # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
        #     iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
        #     weights = iou * scores[None]  # box weights
        #     x[i, :4] = th.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
        #     if redundant:
        #         i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        # if mps:
        #     output[xi] = output[xi].to(device)
        # if (time.time() - t) > time_limit:
        #     LOGGER.warning(f'WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded')
        #     break  # time limit exceeded

  return output



# class ModelWrapper(th.nn.Module):
#   def __init__(self, path):
#     super(ModelWrapper, self).__init__()
#     extra_files = {'config.txt' : ''}
#     self.backbone_model = th.jit.load(path, map_location=th.device('cpu'), _extra_files=extra_files)
#     self.config = json.loads(extra_files['config.txt' ].decode('utf-8'))
#     return

#   def forward(self, x):
#     return self.backbone_model(x)


class Y5(th.nn.Module):
  def __init__(self, path, dev, topk=False):
    super(Y5, self).__init__()
    # self.backbone_model = ModelWrapper(path=path)
    # self.config = self.backbone_model.config
    extra_files = {'config.txt' : ''}
    self.backbone_model = th.jit.load(path, map_location=dev, _extra_files=extra_files)
    self.config = json.loads(extra_files['config.txt' ].decode('utf-8'))
    self.y5_nms = y5_nms_topk if topk else y5_nms
    return

  def forward(self, inputs):
    th_x = self.backbone_model(inputs)
    th_x = th_x[0]
    th_x = self.y5_nms(th_x)
    return th_x

class Y8(th.nn.Module):

  def __init__(
    self,
    model,
    dev : th.device,
    topk : bool = False,
    backend_type : BackendType = BackendType.TORCH
  ):
    """
    Constructs an Y8 module.

    Parameters:
      model - either a path on disk (string) to the torchscript model or a tuple
        containing the YOLO model and config dict.
        If a torchscript path is provided this will be exportable only to Torchscript.
        Otherwise, the model will be exportable to any backend of Torchscript, ONNX, OpenVino
        and TensorRT depending on the backend_type parameter.
      dev - the torch device that should be used.
      topk - True if we should use the topk NMS (top 6)
      backend_type - The backend type this model is build for
    """
    super(Y8, self).__init__()
    extra_files = {'config.txt' : ''}
    if isinstance(model, str):
      self.backbone_model = th.jit.load(model, map_location=dev, _extra_files=extra_files)
      self.config = json.loads(extra_files['config.txt' ].decode('utf-8'))
      if backend_type != BackendType.TORCH:
        raise ValueError("Non-torch backends should get the model as tuple of th.nn.Module and config!")
    elif isinstance(model, tuple):
      self.backbone_model, self.config = model
    else:
      raise ValueError("Unexpected value for model")

    self.y8_nms = y8_nms_topk if topk else y8_nms
    if backend_type == BackendType.TENSORRT:
      self.y8_nms = y8_nms_trt_efficient
    elif backend_type in [BackendType.ONNX, BackendType.OpenVINO]:
      # OpenVINO as ONNX use the same model for distribution
      # (in onnx format).
      self.y8_nms = y8_nms_onnx
    return

  def forward(self, inputs):
    th_x = self.backbone_model(inputs)
    if isinstance(th_x, tuple):
      # The non-exported torch Y8 model will return a tuple here,
      # but we're only interested in the first value (the detections).
      th_x = th_x[0]
    th_x = self.y8_nms(th_x)
    return th_x


def predict(m, data, m_name, config, log, timing=False):
  if timing:
    log.start_timer(m_name)
  with th.no_grad():
    th_preds = m(data)
  if not config.get('includes_nms', False):
    if 'v5' in config.get('model',''):
      preds = th_preds[0]
      if timing:
        log.start_timer(m_name + '_nms')
      pred_nms = y5_nms_orig(
         prediction=preds,
       )

      if timing:
        log.start_timer(m_name + '_nms_cpu')
      pred_nms_cpu = [x.cpu().numpy() for x in pred_nms]
      if timing:
        log.stop_timer(m_name + '_nms_cpu')

      if timing:
        log.stop_timer(m_name + '_nms')
    else:
      if timing:
        log.start_timer(m_name + '_nms')
      pred_nms = y8_nms_orig(
         prediction=th_preds,
      )

      if timing:
        log.start_timer(m_name + '_nms_cpu')
      pred_nms_cpu = [x.cpu().numpy() for x in pred_nms]
      if timing:
        log.stop_timer(m_name + '_nms_cpu')

      if timing:
        log.stop_timer(m_name + '_nms')
    #endif y5 vs y8

  else:
    if timing:
      log.start_timer(m_name + '_cpu')
    th_preds, th_n_det = th_preds
    pred_nms_cpu = [x[:th_n_det[i]].cpu().numpy() for i, x in enumerate(th_preds)]
    if timing:
      log.stop_timer(m_name  + '_cpu')
  #end requires nms
  if timing:
    log.stop_timer(m_name)
  return pred_nms_cpu



#local dependenciess
from naeural_core.local_libraries.vision.benchmark.plugins.pb_model import PbModel

_CONFIG_ED4_None_BS1 = {
    'GRAPH': 'effdet4_None/efficientdetd4_b1_None/efficientdet-d4_frozen.pb',
    'INPUT_TENSORS': ['image_arrays:0', 'min_score_thresh:0', 'iou_threshold_thresh:0'],
    'OUTPUT_TENSORS': ['detections:0'],
    'IOU_THRESHOLD': 0.3,
    'MEMORY_FRACTION': 0.3,
    'MODEL_THRESHOLD': 0.3
  }

_CONFIG_ED4_None_BS2 = {
    'GRAPH': 'effdet4_None/efficientdetd4_b2_None/efficientdet-d4_frozen.pb',
    'INPUT_TENSORS': ['image_arrays:0', 'min_score_thresh:0', 'iou_threshold_thresh:0'],
    'OUTPUT_TENSORS': ['detections:0'],
    'IOU_THRESHOLD': 0.3,
    'MEMORY_FRACTION': 0.3,
    'MODEL_THRESHOLD': 0.3
  }

_CONFIG_ED4_None_BS4 = {
    'GRAPH': 'effdet4_None/efficientdetd4_b4_None/efficientdet-d4_frozen.pb',
    'INPUT_TENSORS': ['image_arrays:0', 'min_score_thresh:0', 'iou_threshold_thresh:0'],
    'OUTPUT_TENSORS': ['detections:0'],
    'IOU_THRESHOLD': 0.3,
    'MEMORY_FRACTION': 0.3,
    'MODEL_THRESHOLD': 0.3
  }

_CONFIG_ED4_None_BS8 = {
    'GRAPH': 'effdet4_None/efficientdetd4_b8_None/efficientdet-d4_frozen.pb',
    'INPUT_TENSORS': ['image_arrays:0', 'min_score_thresh:0', 'iou_threshold_thresh:0'],
    'OUTPUT_TENSORS': ['detections:0'],
    'IOU_THRESHOLD': 0.3,
    'MEMORY_FRACTION': 0.3,
    'MODEL_THRESHOLD': 0.3
  }

class ED4_None_BS1(PbModel):
  def __init__(self, **kwargs):
    config_model = kwargs.pop('config_model', _CONFIG_ED4_None_BS1)
    super().__init__(config_model=config_model, **kwargs)
    self.model_threshold = self.config_model['MODEL_THRESHOLD']
    self.iou_threshold = self.config_model['IOU_THRESHOLD']
    return
  
  def _process_input(self, inputs, **kwargs):    
    dct_inputs = {
      self.input_tensor_names[0] : inputs,
      self.input_tensor_names[1] : self.model_threshold,
      self.input_tensor_names[2] : self.iou_threshold
    }
    return dct_inputs


class ED4_None_BS2(ED4_None_BS1):
  def __init__(self, **kwargs):
    super().__init__(config_model=_CONFIG_ED4_None_BS2, **kwargs)
    return
  

class ED4_None_BS4(ED4_None_BS1):
  def __init__(self, **kwargs):
    super().__init__(config_model=_CONFIG_ED4_None_BS4, **kwargs)
    return
  
  
class ED4_None_BS8(ED4_None_BS1):
  def __init__(self, **kwargs):
    super().__init__(config_model=_CONFIG_ED4_None_BS8, **kwargs)
    return




"""
 "HEAVY_OPS_CONFIG" : {
    "ACTIVE_COMM_ASYNC" : [
      'send_mail',
      'debug_save',
      'save_image_dataset',
      'a_dummy'
    ],
    
    "ACTIVE_ON_COMM_THREAD" : [
      'image_compression',
    ]
  }
  
"""

DEFAULT_HEAVY_OPS_CONFIG = {
  "ACTIVE_COMM_ASYNC" : [
    "send_mail",
    "save_image_dataset",
  ],

  "ACTIVE_ON_COMM_THREAD" : [
  ]     
}

from naeural_core.manager import Manager
from naeural_core import constants as ct


class HeavyOpsManager(Manager):

  def __init__(self, log, shmem, **kwargs):
    self.shmem = shmem
    self._dct_ops = None
    super(HeavyOpsManager, self).__init__(log=log, prefix_log='[HOPSM]', **kwargs)
    return

  def startup(self):
    super().startup()
    self.config_data = self.config_data.get('HEAVY_OPS_CONFIG', DEFAULT_HEAVY_OPS_CONFIG)
    self._dct_ops = self._dct_subalterns
    # 1st category as heavy ops that run on separate individual threads without (usually)
    # affecting inplace the payload
    for operation_name in self.config_data.get("ACTIVE_COMM_ASYNC", []):
      self.create_heavy_operation(operation_name, comm_async=True)
      
    # 2nd category are heavy ops that do not use separate thread and run on comms thread
    # this second category is usually for plugins that change inplace the payload
    for operation_name in self.config_data.get("ACTIVE_ON_COMM_THREAD", []):
      self.create_heavy_operation(operation_name, comm_async=False)
    return

  def _get_plugin_class(self, name):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_HEAVY_OPS_PLUGINS,
      name=name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_HEAVY_OPS_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_HEAVY_OPS_PLUGINS,
      safety_check=True, # perform safety check           
    )

    if _class_def is None:
      msg = "Error loading heavy_ops plugin '{}'".format(name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info="No code/script defined for heavy_ops plugin '{}' in {}".format(name, ct.PLUGIN_SEARCH.LOC_HEAVY_OPS_PLUGINS)
      )
    #endif

    return _class_def, _class_config

  def create_heavy_operation(self, name, comm_async):
    _cls, _config = self._get_plugin_class(name)

    try:
      op = _cls(log=self.log, shmem=self.shmem, config=_config, comm_async=comm_async)
      self._dct_ops[name] = op
    except Exception as exc:
      msg = "Exception '{}' when initializing heavy_ops plugin {}".format(exc, name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        autocomplete_info=True
      )
    #end try-except
    return

  def run_all_comm_async(self, msg):
    for name, op in self._dct_ops.items():
      if not op.comm_async:
        continue
      op.process_payload(msg)

    return

  def run_all_on_comm_thread(self, msg):
    for name, op in self._dct_ops.items():
      if op.comm_async:
        continue
      op.process_payload(msg)

    return


# -*- coding: utf-8 -*-
import numpy as np
import os
import sys
import torch as th
import torchvision as tv
import json
import cv2

from naeural_core import Logger
from decentra_vision.draw_utils import DrawUtils
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad
from plugins.serving.architectures.y5.general import scale_coords

from naeural_core.xperimental.th_y8.utils import predict

USE_LOCAL_MODELS = False
MODELS_TORCHSCRIPT = [
  {
    'pts': "C:/repos/edge-node/_local_cache/_models/20230723_y8l_nms.ths",
    'model': 'y8l'
  },
]

if __name__ == "__main__":
  TEST = True
  TEST_FP16 = True
  TEST_ONLY_FP16 = True
  FP_VALUES = [True] if TEST_ONLY_FP16 else [False, True] if TEST_FP16 else [False]
  SHOW = True
  N_TESTS = 1
  USE_IMAGES_FROM_FOLDER = True
  DEVICE = 'cuda'
  dev = th.device(DEVICE)

  log = Logger('Y8', base_folder='.', app_folder='_local_cache')
  log.P("Using {}".format(dev))
  models_folder = log.get_models_folder()
  folder = os.path.split(__file__)[0]
  img_names = [
    # 'bus.jpg',
    # 'faces9.jpg',
    # 'faces21.jpg',
    # 'img.png',
    # '688220.png',
    # 'bmw_man3.png',
    # 'faces3.jpg',
    # 'LP3.jpg',
    # 'pic1_crop.jpg'
    "C:/resources/test_lpr/25mai/6cf0d612-617a-4411-97ae-40910733f5ff_O_0.jpg",
    "C:/resources/test_lpr/25mai/dbd5e154-47fd-404b-8b21-871937d5c7db_O_0.jpg",
    "C:/resources/test_lpr/25mai/c3fc83c9-005f-45c3-bfb5-837637ec47e4_O_0.jpg",
    "C:/resources/test_lpr/25mai/bd225a82-54e7-4bfa-8776-636a519cf907_O_0.jpg",

    # "C:/Users/bleot/Dropbox/DATA/_vapor_data/__tests/LPLR/test_lpr/25mai/6cf0d612-617a-4411-97ae-40910733f5ff_O_0.jpg",
    # "C:/Users/bleot/Dropbox/DATA/_vapor_data/__tests/LPLR/test_lpr/25mai/dbd5e154-47fd-404b-8b21-871937d5c7db_O_0.jpg",
    # "C:/Users/bleot/Dropbox/DATA/_vapor_data/__tests/LPLR/test_lpr/25mai/c3fc83c9-005f-45c3-bfb5-837637ec47e4_O_0.jpg",
    # "C:/Users/bleot/Dropbox/DATA/_vapor_data/__tests/LPLR/test_lpr/25mai/bd225a82-54e7-4bfa-8776-636a519cf907_O_0.jpg",
    "_local_cache/_data/Dataset_LPLR_v5.2.2_for_yolo/images/2.Dev/Cars_with_LP/Complexity_2/4.GTS/dataset_builder_1__DATASET_BUILDER_01__DATASET_BUILDER_603.jpg"
  ]

  # assert img is not None
  if USE_IMAGES_FROM_FOLDER:
    # img_dir = r'C:\repos\edge-node\_local_cache\_output\no_human_detected_PER CV 2-2'
    img_dir = r'C:\repos\edge-node\_local_cache\_output\no_human_detected_PER CV 2-1'

    origs = [
      cv2.imread(os.path.join(img_dir, x))
      for x in os.listdir(img_dir)
      if x.endswith('.jpg') or x.endswith('.png')
    ]
  else:
    origs = [
      # cv2.imread(os.path.join(folder, im_name))
      cv2.imread(im_name)
      for im_name in img_names
    ]
  images = origs
  # images = [
  #   np.ascontiguousarray(img[:,:,::-1])
  #   for img in origs
  # ]

  # models_folder = "C:/repos/edge-node/yn_lpd"
  models_folder = "C:/repos/edge-node/core/xperimental/lpd/traces"
  if USE_LOCAL_MODELS:
    models_for_test = [
      {
        'pts': os.path.join(models_folder, x),
        'model': (
          x[x.find('lpd'):] if 'lpd' in x else x[x.find('y'):]
        ).split('.')[0]
      } for x in os.listdir(models_folder) if ('.ths' in x or '.torchscript' in x) and ('_nms' not in x) and ('y' in x)
    ]
  else:
    models_for_test = MODELS_TORCHSCRIPT

  dev = th.device('cuda')
  if TEST:
    painter = DrawUtils(log=log)

    log.P(
      f"Testing {len(models_for_test)} models{' with FP16' if TEST_FP16 else ''} with {N_TESTS} runs (+{20} for warmup) each on {DEVICE}")
    log.P("Testing on :\n   {}".format("\n   ".join([m['pts'] for m in models_for_test])))
    dct_results = {}
    for m in models_for_test:
      for use_fp16 in FP_VALUES:
        fn_path = m['pts']
        extra_files = {'config.txt': ''}
        model = th.jit.load(
          f=fn_path,
          map_location=dev,
          _extra_files=extra_files,
        )
        if use_fp16:
          model.half()
        # endif use_fp16
        log.P("Done loading model on device {}".format(dev))
        config = json.loads(extra_files['config.txt'].decode('utf-8'))
        imgsz = config.get('imgsz', config.get('shape'))
        includes_nms = config.get('includes_nms')
        includes_topk = config.get('includes_topk')
        model_name = m['model']
        model_result_name = (model_name + '_fp16') if use_fp16 else model_name
        if model_result_name not in dct_results:
          dct_results[model_result_name] = {}
        is_y5 = 'y5' in model_name
        if includes_topk:
          model_name = model_result_name + '_topk'
        else:
          model_name = model_result_name + ('_inclnms' if includes_nms else '')
        log.P("Model {}: {}, {}:".format(model_name, imgsz, fn_path, ), color='m')
        maxl = max([len(k) for k in config])
        for k, v in config.items():
          if not (isinstance(v, dict) and len(v) > 5):
            log.P("  {}{}".format(k + ':' + " " * (maxl - len(k) + 1), v), color='m')

        h, w = imgsz[-2:]
        log.P("  Resizing from {} to {}".format([x.shape for x in images], (h, w)))
        class_names = config['names']
        results = th_resize_with_pad(
          img=images,
          h=h,
          w=w,
          device=dev,
          normalize=True,
          return_original=False,
          half=use_fp16
        )
        if len(results) < 3:
          prep_inputs, lst_original_shapes = results
        else:
          prep_inputs, lst_original_shapes, lst_original_images = results

        # warmup
        log.P("  Warming up...")
        for _ in range(20):
          print('.', flush=True, end='')
          pred_nms_cpu = predict(model, prep_inputs[:5], model_name, config, log=log, timing=False)
        print('')

        # timing
        log.P("  Predicting...")
        for _ in range(N_TESTS):
          print('.', flush=True, end='')
          pred_nms_cpu = predict(model, prep_inputs, model_name, config, log=log, timing=True)
        print('')

        log.P("  Last preds:\n{}".format(pred_nms_cpu))

        mkey = 'includes_topk' if includes_topk else 'includes_nms' if includes_nms else 'normal'

        dct_results[model_result_name][mkey] = {
          'res': pred_nms_cpu,
          'time': log.get_timer_mean(model_name),
          'name': model_name
        }

        if SHOW:
          log.P(f"  {'[FP16]' if use_fp16 else ''}Showing...")
          for i in range(len(images)):
            # now we have each individual image and we generate all objects
            # what we need to do is to match `second_preds` to image id & then
            # match second clf with each box
            img_bgr = origs[i].copy()
            np_pred_nms_cpu = pred_nms_cpu[i]
            original_shape = lst_original_shapes[i]
            np_pred_nms_cpu[:, :4] = scale_coords(
              img1_shape=(h, w),
              coords=np_pred_nms_cpu[:, :4],
              img0_shape=original_shape,
            ).round()
            lst_inf = []
            for det in np_pred_nms_cpu:
              det = [float(x) for x in det]
              # order is [left, top, right, bottom, proba, class] => [L, T, R, B, P, C, RP1, RC1, RP2, RC2, RP3, RC3]
              L, T, R, B, P, C = det[:6]  # order is [left, top, right, bottom, proba, class]
              label = class_names[str(int(C))]
              img_bgr = painter.draw_detection_box(image=img_bgr, top=int(T), left=int(L), bottom=int(B), right=int(R),
                                                   label=label, prc=P)
            cv2.imshow(fn_path, img_bgr)
            cv2.moveWindow(fn_path, 0, 0)
            pressed_key = cv2.waitKey(0)
            cv2.destroyAllWindows()
            if pressed_key == 27:
              break
          # endfor plot images
        # endif show
      # endfor fp16
    # endfor each model
  # endif test models
  log.show_timers()
  for mn in dct_results:
    if 'normal' not in dct_results[mn] or 'includes_nms' not in dct_results[mn]:
      continue
    a1 = dct_results[mn]['normal']['res']
    t1 = dct_results[mn]['normal']['time']
    a2 = dct_results[mn]['includes_nms']['res']
    t2 = dct_results[mn]['includes_nms']['time']
    # a3 = dct_results[mn]['includes_topk']['res']
    # t3 = dct_results[mn]['includes_topk']['time']
    # a3 = [x[:, :6] for x in a3]

    ok1 = all([np.allclose(a1[i], a2[i]) for i in range(len(a1))])
    # ok2 = all([np.allclose(a1[i], a3[i]) for i in range(len(a1))])

    gain1 = t1 - t2
    rel_gain1 = gain1 / t1
    # gain2 = t1-t3
    log.P(f'Model with NMS {mn} {t1} => {t2}[gain {gain1:.5f}s({rel_gain1:.2f}%)], equal: {ok1}',
          color='r' if not ok1 else 'g')
    # log.P("Model with NMS {} gain {:.5f}s, equal: {}".format(mn, gain1, ok1), color='r' if not ok1 else 'g')
    # log.P("Model with TopK {} gain {:.5f}s, equal: {}".format(mn, gain2, ok2), color='r' if not ok2 else 'g')




# global dependencies
import os
import abc
import cv2
import numpy as np

from collections import defaultdict

# local dependencies
from naeural_core import constants as ct
from naeural_core.business.mixins_base.debug_info_mixin import _DebugInfoMixin

from .base_plugin_biz import BasePluginExecutor
from naeural_core.serving.ai_engines import AI_ENGINES
from naeural_core.business import utils
from naeural_core.utils import nms
from decentra_vision.draw_utils import DrawUtils

from naeural_core.business.mixins_base import _LimitedDataMixin

# TODO: this will be moved or replaced
try:
  from extensions.business.mixins_base.multi_class_tracker_mixin import _MultiClassTrackerMixin
except:
  from naeural_core.business.mixins_base.multi_class_tracker_mixin import _MultiClassTrackerMixin

from naeural_core.business.mixins_libs import _AlertWitnessMixin

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'DEBUG_ON_WITNESS': True,  # if true, the debug info will not be drawn on the image

  # Tracker
  'TRACKING_ENABLED': True,
  'TRACKING_MODE': 0,  # 0 - linear, 1-sort,
  'LINEAR_RESET_MINUTES': 60,
  'LINEAR_MAX_DISTANCE_RATIO': 4,
  'LINEAR_MAX_DIST_SCALE': 1.4,
  'LINEAR_MAX_RELATIVE_DIST': 1.2,
  'LINEAR_CENTER_DIST_WEIGHT': 1,
  'LINEAR_HW_DIST_WEIGHT': 0.8,
  'LINEAR_MAX_AGE': 4,
  'SORT_MIN_IOU': 0,
  'SORT_MAX_AGE': 3,
  'SORT_MIN_HITS': 1,
  
  "COLOR_TAGGING": False,  

  'PRC_INTERSECT': 0.5,
  'TRUSTED_PRC': 0.65,
  'DEBUG_OBJECTS': False,
  'DEBUG_OBJECTS_PATHS': False,
  'META_TYPE_MAPPING': {},
  'META_NMS_IOU_THR': 0.6,


  'RUN_WITHOUT_IMAGE': False,
  'ALLOW_EMPTY_INPUTS': False,  # if this is set to true the on-idle will be triggered continously the process

  'DEBUG_INFO_ACTIVE_ALERTER_TIME': 30,  # seconds, used to reduce the data dumped by alerters
  'DEBUG_SHOW_DEFAULT_ALERTER': True,

  'ALPHA_OUTER_ZONE': True,

  'ALPHA_OUTER_ZONE_INTENSITY': 170,

  'SAVE_TARGET_ZONE': True,

  'THREAD_SAFE_DRAWING': True,

  'POINTS': [],
  'OBJECT_TYPE': [],

  'ADD_ORIGINAL_IMAGE': False,

  'DEMO_MODE_TIME_LIMIT': 120,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],

    'ADD_ORIGINAL_IMAGE': {
      'TYPE': 'bool',
      'DESCRIPTION': "Add original image to payloads as 'IMG_ORIG' beside the usual witness 'IMG' key",
    },

    'META_TYPE_MAPPING': {
      'TYPE': 'dict',
      'DESCRIPTION': 'dict(key: list())  Mapping dictionary that defines each meta type as a list of types. '
                     'If an meta type is set for a given type, the object tracker will use the said type for aggregation'
    },

    'POINTS': {
      'TYPE': 'list',
      # No need for `'MIN_LEN' : 4,` as there is validate_coords
    },

    'RUN_WITHOUT_IMAGE': {
      'DESCRIPTION': "The cv plugin instance will run its inner loop even if there is no image from upstream",
      'TYPE': 'bool'
    },

    'DEBUG_OBJECTS_PATHS': {
      'TYPE': 'bool',
      'DESCRIPTION': 'Draw the paths history of all objects. Works only with debug objects true'
    },

    ### Tracking ###
    'TRACKING_ENABLED': {
      'TYPE': 'bool',
      'DESCRIPTION': 'Flag for enabling tracking. Default `True`'
    },

    'TRACKING_MODE': {
      'TYPE': 'int',
      'DESCRIPTION': 'What tracking algorithm to be used. 1 - linear, 2 - sort. Default 1'
    },

    'LINEAR_RESET_MINUTES': {
      'TYPE': 'int',
      'DESCRIPTION': 'The maximum life of an tracked object. After the number of minutes passes the object is given a '
                     'new id. Only for linear tracking mode'
    },

    'LINEAR_MAX_DISTANCE_RATIO': {
      'TYPE': 'float',
      'DESCRIPTION': 'Parameter used to heuristically compute the maximum allowed distance between objects to have the'
                     'same id. Formula is: max_distance = sqrt(image_h * image_w) / LINEAR_MAX_DISTANCE_RATIO. '
                     'Only for linear tracking mode'
    },

    'LINEAR_MAX_RELATIVE_DIST': {
      'TYPE': 'float',
      'DESCRIPTION': 'Parameter used to heuristically compute the maximum allowed distance between 2 consecutive '
                     'appearances of the same object. The formula is: '
                     'max_relative_distance = dist / max(h1, w1, h2, w2), where dist is the distance between the '
                     'centroid of a detection in the current frame and the centroid of the last appearance of the '
                     'checked track_id and (h1, w1), (h2, w2) are the heights and widths of the current detection '
                     'and the candidate\'s last appearance.'
    },

    'LINEAR_MAX_DIST_SCALE': {
      'TYPE': 'float',
      'DESCRIPTION': 'Parameter used to compute maximum distance per distance type.  Formula is :'
                     'distance_type_max = max_distance * dist_weight / (hw_dist_weight + center_dist_weight) * max_dist_scale.'
                     'Should have values between 1 and 2. Only for linear tracking mode'
    },

    'LINEAR_CENTER_DIST_WEIGHT': {
      'TYPE': 'float',
      'DESCRIPTION': 'The weight of the centroid distance in computing the total distance. Only for linear tracking mode'
    },

    'LINEAR_HW_DIST_WEIGHT': {
      'TYPE': 'float',
      'DESCRIPTION': 'The weight of the sizes distance in computing the total distance. Only for linear tracking mode'
    },
    'LINEAR_MAX_AGE': {
      'TYPE': 'int',
      'DESCRIPTION': 'Number of untracked frames until the object id is no loger tracked. Only for linear tracking mode'
    },

    'SORT_MIN_IOU': {
      'TYPE': 'float',
      'DESCRIPTION': 'The minimum IOU between objects from different frames to be considered tracking candidates'
    },
    'SORT_MAX_AGE': {
      'TYPE': 'int',
      'DESCRIPTION': 'Number of untracked frames until the object id is no loger tracked. Only for sort tracking mode'
    },
    'SORT_MIN_HITS': {
      'TYPE': 'int',
      'DESCRIPTION': 'Number of identifications to be assigned an id.'
    }
  },
}


class CVPluginExecutor(BasePluginExecutor,
                       _LimitedDataMixin,
                       _MultiClassTrackerMixin,
                       _DebugInfoMixin,
                       _AlertWitnessMixin
                       ):
  __metaclass__ = abc.ABCMeta

  CONFIG = _CONFIG

  def __init__(self, **kwargs):

    self._irregular_target_area = None
    self._coords_type = None
    self._top, self._left, self._bottom, self._right = None, None, None, None
    self.saved_zone = False
    self.demo_mode_time = None

    self.__last_witness_images = None
    self.__last_witness_count = 0

    super(CVPluginExecutor, self).__init__(**kwargs)
    return

  def set_output_image(self, img):
    self.default_image = img
    return

  def set_image(self, img):
    return self.set_output_image(img)

  def startup(self):
    super().startup()
    self._painter = DrawUtils(log=self.log, timers_section=self._timers_section)
    return

  def _get_detector_ai_engine(self):
    configured_ai_engines = self.cfg_ai_engine
    if type(configured_ai_engines) != list:
      configured_ai_engines = [configured_ai_engines]

    # TODO: FIND A BETTER METHOD
    detectors = [x for x in AI_ENGINES.keys() if 'detector' in x]

    for ai_engine in configured_ai_engines:
      if ai_engine in detectors:
        return ai_engine

    return None

  def _get_detector_ai_engines(self):
    """
    Get a list of all detector engines that are configured for this plugin
    """
    configured_ai_engines = self.cfg_ai_engine
    if type(configured_ai_engines) != list:
      configured_ai_engines = [configured_ai_engines]

    # TODO: FIND A BETTER METHOD
    configured_detectors = [x for x in AI_ENGINES.keys() if 'detector' in x and x in configured_ai_engines]

    return configured_detectors

  @property
  def _type_to_meta_type_map(self):
    return {type: meta_type for meta_type, types in self.cfg_meta_type_mapping.items() for type in types}

  def _get_object_areas(self, filter_object_type):
    # Returns object areas / total image are
    detector_ai_engine = self._get_detector_ai_engine()
    if detector_ai_engine is None:
      return []

    inferences = self.dataapi_inferences()[detector_ai_engine][0]

    if filter_object_type:
      inferences = [inference for inference in inferences if inference['TYPE'] == filter_object_type]

    img_area = self.dataapi_image().shape[0] * self.dataapi_image().shape[1]

    proportions = [
      (inference['TLBR_POS'][3] - inference['TLBR_POS'][1]) *
      (inference['TLBR_POS'][2] - inference['TLBR_POS'][0]) /
      img_area
      for inference in inferences
    ]
    return proportions

  def _get_tlbr(self):
    points = self.cfg_points
    if len(points) == 0:
      self._coords_type = ct.COORDS_NONE
    elif isinstance(points[0], int):
      self._coords_type = ct.COORDS_TLBR
    elif isinstance(points[0], list):
      self._coords_type = ct.COORDS_POINTS
    else:
      self._coords_type = ct.COORDS_NONE
    # endif

    self._irregular_target_area = (self._coords_type == ct.COORDS_POINTS)

    if self._coords_type == ct.COORDS_POINTS:
      self._top, self._left, self._bottom, self._right = utils.convert_points_to_tlbr(
        points=self.cfg_points
      )
    elif self._coords_type == ct.COORDS_TLBR:
      self._top, self._left, self._bottom, self._right = self.cfg_points

    return

  def _keep_only_intersected_objects(self, dct_positional_plugin_inference):
    prc_intersect = self.cfg_prc_intersect

    dct_intersected = {}
    if self._coords_type == ct.COORDS_POINTS:
      func = utils.keep_boxes_inside_irregular_target_area
      target_area = self.cfg_points
    else:
      func = utils.keep_boxes_inside_regular_target_area
      target_area = [self._top, self._left, self._bottom, self._right]

    for model, lst_2d in dct_positional_plugin_inference.items():
      new_lst_2d = []
      for lst in lst_2d:
        new_lst, intersect = [], []
        if len(lst) > 0:
          new_lst, intersect = func(
            lst_box=lst,
            target_area=target_area,
            prc_intersect=prc_intersect
          )
        # endif

        for idx, x in enumerate(new_lst):
          x[ct.PRC_INTERSECT] = intersect[idx]

        new_lst_2d.append(new_lst)
      # endfor
      dct_intersected[model] = new_lst_2d
    # endfor

    return dct_intersected

  def _filter_confidence_threshold(self, dct_inference):
    dct = {}
    conf_thr = self._get_confidence_threshold()
    if conf_thr is None or conf_thr == 0:
      dct = dct_inference
    else:
      for model, lst_2d in dct_inference.items():
        new_lst_2d = []
        for lst in lst_2d:
          new_lst = []
          for elem in lst:
            if isinstance(elem, dict):
              prob_prc = elem.get(ct.PROB_PRC, None)
              if isinstance(prob_prc, (tuple, list, type(None))):
                new_lst.append(elem)
              elif prob_prc >= conf_thr:
                elem['IS_TRUSTED'] = int(prob_prc > self.cfg_trusted_prc)
                new_lst.append(elem)
              # endif
            else:
              new_lst.append(elem)
            # endif
          # endfor - elem
          new_lst_2d.append(new_lst)
        # endfor - lst
        dct[model] = new_lst_2d
      # endfor dct_inference

    return dct

  def _filter_object_types(self, dct_inference):
    dct = {}
    for model, lst_2d in dct_inference.items():
      new_lst_2d = self._keep_only_plugin_object_types(lst_2d)
      dct[model] = new_lst_2d
    return dct

  def _keep_only_plugin_object_types(self, lst_2d):
    if len(self.cfg_object_type) > 0:
      l_filtered = []
      for lst in lst_2d:
        l_filtered.append([x for x in lst if x[ct.TYPE] in self.cfg_object_type])
    else:
      l_filtered = lst_2d
    return l_filtered

  def _get_generic_positional_objects(self, dct_inference):
    """
    Returns both spatial (with TLBR) objects as well as non-spatial ones

    Parameters
    ----------
    dct_inference : dict
      Upstream inference dict.

    Returns
    -------
    dct_generic : dict
      Non-TLBR objects.
    dct_positional : dict
      TLBR objects.

    """
    dct_generic, dct_positional = {}, {}
    for model, lst_2d in dct_inference.items():
      if len(lst_2d) > 0 and len(lst_2d[0]) > 0:
        if ct.TLBR_POS in lst_2d[0][0]:
          dct_positional[model] = lst_2d
        else:
          dct_generic[model] = lst_2d
      else:
        dct_generic[model] = lst_2d
      # endif
    # endfor

    return dct_generic, dct_positional

  def get_debug_objects_summary(self, debug_objects):
    objs = debug_objects
    if isinstance(objs, list) and len(objs) > 0:
      dct_res = defaultdict(lambda: 0)
      for obj in objs:
        dct_res[obj[ct.TYPE]] += 1
      return dict(dct_res)
    return "Nothing"

  def _draw_and_label_target_area(self, img_witness, irregular_target_area, cfg_instance_id, points, tlbr):
    if irregular_target_area:
      img_witness = self._draw_target_area(
        image=img_witness,
        name=cfg_instance_id,
        irregular_target_area=irregular_target_area,
        points=points,
        tlbr=tlbr
      )

      if self.cfg_alpha_outer_zone:
        img_witness = self._painter.alpha_outer_poly_area(
          image=img_witness,
          points=points,
          intensity=self.cfg_alpha_outer_zone_intensity,
        )
    else:
      MARGIN = 10
      top, left, bottom, right = tlbr
      _h, _w = img_witness.shape[:2]
      if left > MARGIN or top > MARGIN or bottom < (_h - MARGIN) or right < (_w - MARGIN):
        img_witness = self._draw_target_area(
          image=img_witness,
          name=cfg_instance_id,
          irregular_target_area=irregular_target_area,
          points=points,
          tlbr=tlbr
        )

        if self.cfg_alpha_outer_zone:
          img_witness = self._painter.alpha_outer_area(
            image=img_witness,
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            intensity=self.cfg_alpha_outer_zone_intensity,
          )
      # endif
    # endif
    return img_witness

  def _draw_target_area(self, image, name, irregular_target_area, points, tlbr, blur=False, use_rectangle_target=True):
    _top, _left, _bottom, _right = tlbr
    top, left = None, None
    if irregular_target_area:
      # draw polygon
      img = self._painter.polygon(
        image=image,
        pts=points,
        color=ct.GREEN,
        thickness=3
      )

      # extract positions for name
      arr = np.array(points)
      top = arr[:, 1].min()
      argmin = arr[:, 1].argmin()
      left = points[argmin][0]
      top = top if top > 10 else 10
      left = left if left > 10 else 10

      if blur:
        img = self._painter.alpha_inner_poly_area(
          image=image,
          points=points
        )
    else:
      # draw rectangle
      if use_rectangle_target:
        fn = self._painter.rectangle_target
      else:
        fn = self._painter.rectangle

      img = fn(
        image=image,
        pt1=(_left, _top),
        pt2=(_right, _bottom),
        color=ct.GREEN,
        thickness=3
      )

      # extract positions for name
      top = _top if _top > 0 else 10
      left = _left if _left > 0 else 10

      if blur:
        img = self._painter.alpha_inner_area(
          image=image,
          top=_top,
          left=_left,
          bottom=_bottom,
          right=_right
        )
      # endif
    # endif

    if name:
      img = self._painter.alpha_text_rectangle(
        image=img,
        top=top,
        left=left,
        text=name,
        color=ct.DARK_GREEN
      )
    # endif
    return img

  def on_command(self, data, get_last_witness=None, **kwargs):
    """
    Called when the instance receives new INSTANCE_COMMAND

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    Observation
    -----------

      For "get-last-witness" command we have the following command structure:
      ```
        {
          'NAME': '<name of pipeline>', 
          'INSTANCE_ID': '<name of instance>', 
          'SIGNATURE': '<name of signature>', 
          'INSTANCE_CONFIG': 
          {
            "REQUESTED": "example-key-from-a-backend-appd", 
            "REQUEST_ID": "e6b3abfd-8001-45fd-b0d7-2a6fea6b8949",
            "COMMAND_PARAMS": 
            {
              "GET_LAST_WITNESS": true
            }
          }
        }
      ```
      all the `COMMAND_PARAMS` are optional, but if present, they must be a dict and will be transformed
      into kwargs for the `_on_command` method.

    """
    if (isinstance(data, str) and data.upper == 'GET_LAST_WITNESS') or get_last_witness:
      self.P("Received `GET_LAST_WITNESS` request command. Running default callback.)")
      images = self.__last_witness_images
      self.add_payload_by_fields(
        img=images,
        last_witness_count=self.__last_witness_count,
        command_params=data,
      )
    return

  # Section A - methods overwritten from parenn
  # When reimplementing, pay double attention if you want to overwrite the entire functionality;
  # otherwise, first line should be parent_result = super().method_name(...)
  if True:
    def high_level_execution_chain(self):
      if (not self.cfg_run_without_image and len(self.dataapi_images()) == 0) and not self.cfg_allow_empty_inputs:
        # cancel execution if and only if we must have image and no image is avail from upstream
        # and allow_empty inputs is false
        # due to `allow_empty_inputs` image check is half-redundant
        return
      # endif

      if len(self.dataapi_images()) > 0:
        self.__last_witness_images = self.dataapi_images_as_list()
        self.__last_witness_count += 1
      # endif
      super().high_level_execution_chain()
      return

    def _update_instance_config(self):
      super()._update_instance_config()
      self._get_tlbr()
      return

    def _prepare_payload_for_testing_registration(self, payload):
      super()._prepare_payload_for_testing_registration(payload)
      payload['_T_FRAME_CURRENT'] = self.limited_data_frame_current
      payload['_T_FPS'] = self.limited_data_fps
      payload['_T_SECONDS_ELAPSED'] = self.limited_data_seconds_elapsed
      payload['_T_CRT_TIME'] = self.limited_data_crt_time
      return payload

    def _prepare_payload_for_generating_testing_results(self, payload):
      super()._prepare_payload_for_generating_testing_results(payload)
      if self._limited_data_finished:
        _payload = {
          'Y_HAT': self._testing_manager.tester.y_hat,
          'EXCEPTIONS': self._testing_manager.tester.exceptions,
          'PROCESSING_SECONDS': self._last_process_time - self._first_process_time,
          'MOVIE_SECONDS': self.limited_data_duration
        }
      return _payload

    def _print_warnings(self):
      super()._print_warnings()
      if self.cfg_debug_objects:
        self.P("Plugin instance draws ALL objects for DEBUG purposes.", color='y')

      if not self.cfg_thread_safe_drawing:
        self.P("Plugin instance NOT use thread-safe for witness drawing.", color='y')

      if self.cfg_cancel_witness:
        self.P("Plugin instance does NOT send witness image in payloads.", color='y')
      return

    def _prepare_debug_save_payload(self):
      # this method is "connected" with "debug_save" heavy ops plugin

      if self._payload is None:
        return

      self._payload['_H_ORIGINAL_IMAGE'] = self.dataapi_image()
      self._payload['_H_RELATIVE_PATH'] = self.instance_relative_path
      # TODO: maybe remove? we have cfg_debug_save_payload now
      self._payload['_H_ARCHIVE_M'] = self.cfg_debug_save_archive_m
      self._payload['_H_UPLOAD'] = self.cfg_debug_save_upload
      ####
      return

    # TODO: maybe remove
    def _add_plugin_identifiers_to_payload(self, payload):
      super()._add_plugin_identifiers_to_payload(payload)
      vars(payload)['_P_' + ct.CONFIDENCE_THRESHOLD] = self._get_confidence_threshold()

      if len(self.cfg_points) > 0:
        vars(payload)['_P_' + ct.LOCATION] = self.cfg_points
      if len(self.cfg_object_type) > 0:
        vars(payload)['_P_' + ct.OBJECT_TYPE] = self.cfg_object_type
      return

    def _draw_object_centroid_history(self, img, inference, color=None, nr_steps=None, draw_orientation=False):
      centroid_history = self._get_object_history(
        object_id=inference[self.consts.TRACK_ID],
        object_type=inference[self.consts.TYPE]
      )
      if nr_steps is not None and nr_steps > 0:
        centroid_history = centroid_history[-nr_steps:]
      # endif

      if color is None:
        color_list = self.sns.color_palette("Paired")
        color_list = [[int(channel * 255) for channel in color] for color in color_list]
        color = color_list[inference[self.consts.TRACK_ID] % len(color_list)]
      # endif color not provided

      centroid_history = [self.np.flip(x) for x in centroid_history]
      img = self._painter.crosses(img, centroid_history, color=color, thickness=2, size_factor=2e-3)
      lines = [(centroid_history[i], centroid_history[i + 1]) for i in range(len(centroid_history) - 1)]
      for line in lines:
        img = self._painter.line(img, line[0], line[1], color=color, thickness=2)
      # endfor lines

      if draw_orientation:
        orientation = self._get_current_orientation(
          object_id=inference['TRACK_ID'],
          object_type=inference['TYPE']
        )
        img = self._painter.arrow(
          img,
          centroid_history[-1],
          (centroid_history[-1][0] + orientation[1], centroid_history[-1][1] + orientation[0]),
          # In tracker axises are reversed
          color=color, thickness=2
        )
      # endif draw_orientation

      return img

    def _draw_object_centroid_histories(self, img, inferences):
      color_list = self.sns.color_palette("Paired")
      color_list = [[int(channel * 255) for channel in color] for color in color_list]
      for inference in inferences:
        color = color_list[inference[self.consts.TRACK_ID] % len(color_list)]
        img = self._draw_object_centroid_history(img=img, inference=inference, color=color)
      # endfor inference
      return img

    def _witness_pre_process(self, img_witness, **kwargs):
      img_witness = super()._witness_pre_process(img_witness, **kwargs)
      lst_plugin_inferences = kwargs.get('lst_plugin_inferences', [])

      # focus main target area
      if self.cfg_simple_witness:
        return img_witness

      draw_all_boxes = self.cfg_debug_objects

      if draw_all_boxes:
        img_witness = self._painter.draw_inference_boxes(
          image=img_witness,
          lst_inf=lst_plugin_inferences
        )

        if self.cfg_debug_objects_paths and self.cfg_tracking_enabled:
          img_witness = self._draw_object_centroid_histories(img_witness, lst_plugin_inferences)
      # endif

      img_witness = self._draw_and_label_target_area(
        img_witness=img_witness,
        irregular_target_area=self._irregular_target_area,
        cfg_instance_id=self.cfg_instance_id,
        points=self.cfg_points,
        tlbr=(self._top, self._left, self._bottom, self._right)
      )
      return img_witness

    def __get_instance_debug_info(self, img_witness):
      dct_input_metadata = self.dataapi_input_metadata() or {}
      dct_stream_metadata = self.dataapi_stream_metadata() or {}
      cap_time = dct_stream_metadata.get('cap_time', "????-??-?? ??:??:??.??????")
      # nod - pipeline - plug - inst, ver, fps:{}, cap:{}, proc:{}
      instance_info = "{}:{}:{}:{}, {}, fps:{} S:{}x{}, W:{}x{}, cap:{}, proc:{}, tz:{}".format(
        self._device_id,
        self._stream_id,
        self._signature,
        self.cfg_instance_id,
        self.system_version,
        dct_input_metadata.get('fps'),
        dct_input_metadata.get('frame_w'),
        dct_input_metadata.get('frame_h'),
        img_witness.shape[1],
        img_witness.shape[0],
        cap_time.split('.')[0].split(' ')[1],
        self.time_to_str().split(' ')[1],
        self.log.timezone
      )

      return instance_info

    def _witness_post_process(self, img_witness, **kwargs):
      img_witness = super()._witness_post_process(img_witness, **kwargs)
      if self.cfg_simple_witness:  # draw image only if the application runs for `demo purposes`
        return img_witness

      if self.cfg_debug_on_witness:
        img_witness = self._painter.draw_text_outer_image(
          image=img_witness,
          text=[self.__get_instance_debug_info(img_witness=img_witness), str(self.get_serving_processes())],
          font=cv2.FONT_HERSHEY_SIMPLEX,
          font_size=0.4 if img_witness.shape[1] > 720 else 0.32,
          thickness=1,
          location="bottom",
        )

        img_witness = self._draw_debug_info_on_witness(img_witness)
      return img_witness

    def get_witness_image(self, img=None,
                          prepare_witness_kwargs=None,
                          pre_process_witness_kwargs=None,
                          draw_witness_image_kwargs=None,
                          post_process_witness_kwargs=None,
                          zone_only=False):
      if img is None and len(self.dataapi_images()) >= 1:
        # if no image provided just take default image received from upstream
        img = self.dataapi_image()

      if zone_only:
        return self.get_witness_image_zone_only(img=img)

      if pre_process_witness_kwargs is None:
        lst_plugin_positional_inferences = self.dataapi_image_plugin_positional_inferences(raise_if_error=False)
        if lst_plugin_positional_inferences is not None:
          pre_process_witness_kwargs = dict(lst_plugin_inferences=lst_plugin_positional_inferences)

      return super().get_witness_image(
        img=img,
        prepare_witness_kwargs=prepare_witness_kwargs,
        pre_process_witness_kwargs=pre_process_witness_kwargs,
        draw_witness_image_kwargs=draw_witness_image_kwargs,
        post_process_witness_kwargs=post_process_witness_kwargs
      )

    def get_witness_image_zone_only(self, img):
      self.start_timer('get_witness_image_zone_only')
      if img is not None:
        if not isinstance(img, np.ndarray):
          self.P(
            "`get_witness_image` called having a non numpy img {}. Returning no witness".format(repr(self), type(img)),
            color='error')
          return

      uses_BGR = False

      img_witness = self._witness_prepare(img)
      if img_witness is not None:
        img_witness = self._witness_pre_process(img_witness)
        uses_BGR = True

      if img_witness is not None:
        img_witness = self._witness_post_process(img_witness)
        # maybe we need to resize the whole thing so that is respects the model input size
        if len(self.cfg_resize_witness) == 2:
          h, w = self.cfg_resize_witness
          img_witness = self.log.center_image2(
            np_src=img_witness,
            target_h=h,
            target_w=w
          )

        # because we flipped channels in `_witness_prepare` to BGR in order to draw using cv2; now we come back to RGB..
        if not uses_BGR:
          img_witness = img_witness[:, :, ::-1]
      # endif

      self.stop_timer('get_witness_image_zone_only')
      return img_witness

    def process_wrapper(self):
      if not self.saved_zone and self.cfg_save_target_zone:
        # this section will force saving the witness in the local cache if the
        # witness was not previously saved.
        if self.dataapi_received_input():
          self.start_timer('save_zone')
          try:
            dir_path = os.path.join(self.log.get_target_folder('data'), 'zone_witnesses')
            save_path = os.path.join(dir_path, self._signature, self.get_stream_id())

            os.makedirs(save_path, exist_ok=True)
            img_witness = self.get_witness_image(zone_only=True)
            if img_witness is not None:
              fn = os.path.join(save_path, f'{self.cfg_instance_id}.jpg')
              cv2.imwrite(
                filename=fn,
                img=img_witness
              )
          except:
            pass
          self.saved_zone = True
          self.stop_timer('save_zone')
        # end if we have input
      # end if we have not previously saved the target zone
      return super(CVPluginExecutor, self).process_wrapper()

    @staticmethod
    def _maybe_process_meta_types(dct_plugin_inference, meta_type_mapping={}, ai_engine='', nms_iou_thr=0.6):
      detector_inferences = dct_plugin_inference.get(ai_engine, [])
      meta_types = list(meta_type_mapping.keys())
      # if we have no meta_types there is no reason to add them
      if len(meta_types) > 0:
        type_to_meta_type_map = {type: meta_type for meta_type, types in meta_type_mapping.items() for type in types}
        types_idx = {meta_type: it for it, meta_type in enumerate(meta_types)}
        other_cnt = len(meta_types) + 100
        for i, lst_2d in enumerate(detector_inferences):
          # top, left, bottom, right, confidence, class
          tlbrcc = []
          for inf in lst_2d:
            # if the current object's type belongs to a meta type we will add the meta_type
            if inf[ct.TYPE] in type_to_meta_type_map:
              current_meta_type = type_to_meta_type_map[inf[ct.TYPE]]
              inf[ct.META_TYPE] = current_meta_type
              type_id = types_idx[current_meta_type]
            else:
              type_id = other_cnt
              other_cnt += 1
            # endif inf[self.consts.TYPE] in self._type_to_meta_type_map
            # we will append the TLBR, confidence and type_id for nms
            tlbrcc.append(
              [
                *inf[ct.TLBR_POS],
                inf[ct.PROB_PRC],
                type_id
              ]
            )
          # endfor all inferences in image
          if len(tlbrcc) > 0:
            # if we have at least one detection we will filter out the overlapping objects with the same meta_type
            kept_objects_indexes = nms.class_non_max_suppression(
              predictions=np.array(tlbrcc),
              iou_threshold=nms_iou_thr
            )
            detector_inferences[i] = [inf for (j, inf) in enumerate(lst_2d) if kept_objects_indexes[j]]
        # endfor all images
      # endif len(meta_types) > 0
      return dct_plugin_inference
    
    
    def __maybe_complete_color_tag(self, dct_all_inferences, dct_images):
      if self.cfg_color_tagging:
        for model, lst_images_inferences in dct_all_inferences.items():
          for idx, lst_image_inferences in enumerate(lst_images_inferences):
            image = dct_images[idx]
            for infer in lst_image_inferences:              
              if infer.get(ct.COLOR_TAG) is None:
                TLBR = infer.get(ct.TLBR_POS, None)
                T, L, B, R = TLBR
                if TLBR is not None:
                  crop = image[T:B, L:R]
                  color_tag = self.get_crop_color(
                    np_rgb_crop=crop,
                  )
                  if color_tag is not None:
                    infer[ct.COLOR_TAG] = color_tag
                  # endif
                # endif
              # endif
            # endfor inferences
          # endfor images
        # endfor ai-engines
      # endif cfg_color_tagging
      return
    

    def _pre_process(self):
      """
      Pre-process upstream data:
        1. Create list with ai-engine/model for each inference
        2. Filter within location with threshold
      """ 
      dct_images = self.dataapi_images()
      if self._coords_type == ct.COORDS_NONE and len(dct_images) >= 1:
        img = self.dataapi_image()
        self._top, self._left, self._bottom, self._right = 0, 0, img.shape[-3], img.shape[-2]

      dct_global_inferences = self.dataapi_images_inferences()
      

      dct_plugin_inference = self._filter_confidence_threshold(dct_global_inferences)
      dct_plugin_inference = self._filter_object_types(dct_plugin_inference)
      dct_plugin_inference = self._maybe_process_meta_types(
        dct_plugin_inference,
        meta_type_mapping=self.cfg_meta_type_mapping,
        ai_engine=self._get_detector_ai_engine(),
        nms_iou_thr=self.cfg_meta_nms_iou_thr
      )

      # call tracker
      dct_plugin_inference = self._track_objects(dct_plugin_inference)

      dct_generic_plugin_inference, \
          dct_positional_plugin_inference = self._get_generic_positional_objects(dct_plugin_inference)

      dct_positional_inst_inference = self._keep_only_intersected_objects(dct_positional_plugin_inference)

      # now we can update/initiate tracking objects time within target zone
      dct_positional_inst_inference = self._track_in_zone_objects(dct_inference=dct_positional_inst_inference)
      # end track objects time in target zone

      self.__maybe_complete_color_tag(dct_positional_inst_inference, dct_images)

      dct_instance_inference = {**dct_generic_plugin_inference, **dct_positional_inst_inference}
      

      pre_process_outputs = {
        'DCT_PLUGIN_INFERENCES': dct_plugin_inference,
        'DCT_INSTANCE_INFERENCES': dct_instance_inference,

        'DCT_PLUGIN_POSITIONAL_INFERENCES': dct_positional_plugin_inference,
      }

      return pre_process_outputs

    def _post_process(self):
      if len(self.dataapi_images()) >= 1:
        lst_global_inferences = self.dataapi_image_global_inferences()
        lst_instance_inferences = self.dataapi_image_instance_inferences()
      else:
        lst_global_inferences, lst_instance_inferences = [], []

      if self._payload is not None:
        if self.alerter_is_new_alert() and ct.ALERT_OBJECTS not in vars(self._payload):
          vars(self._payload)[ct.ALERT_OBJECTS] = lst_instance_inferences
      else:
        # this is a legacy section
        if self.cfg_debug_payloads and self.need_refresh():  # if in "demo mode"
          self._payload = self._create_payload(
            img=self.get_witness_image()
          )
          self._add_alert_info_to_payload(self._payload)
          debug_objects = lst_global_inferences if self.cfg_debug_objects else "Not enabled"
          vars(self._payload)[ct.DEBUG_OBJECTS] = debug_objects
        # endif send "just image" payloads
      # endif

      if self.is_last_data:
        self._limited_data_finished = True

      self._maybe_disable_demo_mode()

      # reset debug info for next iteration
      self.reset_debug_info()

      return
  # endif

  def _maybe_disable_demo_mode(self):
    if self.demo_mode_time is None and self.cfg_demo_mode:
      self.demo_mode_time = self.time_alive
    elif not self.cfg_demo_mode:
      self.demo_mode_time = None

    if self.cfg_demo_mode and self.time_alive - self.demo_mode_time > self.cfg_demo_mode_time_limit:
      self.update_config_data({'DEMO_MODE': False, 'DEBUG_MODE': False})

  def validate_prc_intersect(self):

    prc_intersect = self.config_data.get(ct.PRC_INTERSECT, None)

    if prc_intersect == 0:
      msg = "Please be careful that 'PRC_INTERSECT' is configured to 0 in {}.".format(self.unique_identification)
      self.add_warning(msg)

    return

  def validate_points(self):
    points = self.config_data.get(ct.POINTS, [])

    if len(points) == 0:
      coords = ct.COORDS_NONE
    elif isinstance(points[0], int):
      coords = ct.COORDS_TLBR
    elif isinstance(points[0], list):
      coords = ct.COORDS_POINTS
    else:
      msg = "Cannot deduce coords type given the points {} in instance {}".format(points, self.unique_identification)
      self.add_error(msg)
      return
    # endif

    if coords == ct.COORDS_TLBR:
      if len(points) != 4:
        msg = "'POINTS' bad configured for COORDS=TLBR in instance {}. It should be a list that contains 4 integers".format(
          self.unique_identification)
        self.add_error(msg)
      elif None in points:
        msg = "'POINTS' bad configured for COORDS=TLBR (it contains None) in instance {}".format(
          self.unique_identification)
        self.add_error(msg)
      # endif
    # endif

    if coords == ct.COORDS_POINTS:
      if not all([len(x) == 2 for x in points]):
        msg = "'POINTS' bad configured for COORDS=POINTS in instance {}.".format(self.unique_identification)
        self.add_error(msg)
      elif None in self.log.flatten_2d_list(points):
        msg = "'POINTS' bad configured for COORDS=POINTS (it contains None) in instance {}".format(
          self.unique_identification)
        self.add_error(msg)
      # endif
    # endif

    return



import os
import zipfile
from naeural_core.data.base import DataCaptureThread

_CONFIG = {
  **DataCaptureThread.CONFIG,
  
  'FORCE_DOWNLOAD'          : False,
  'LIVE_FEED'               : False,
  'RECONNECTABLE'           : 'YES',
  
  
  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}


class MinioDatasetDataCapture(DataCaptureThread):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self._file_system_manager = None
    self._heuristic_cap_resolution = None
    self._send_pings = False
    self.checked_dataset = False
    self.dataset_status = "Checking for dataset"
    super(MinioDatasetDataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    # safe to use config now
    self._setup_minio()
    self._heuristic_cap_resolution = 1/10
    return

  @property
  def cfg_dataset_object_name(self):    
    if self.cfg_url is not None:
      # if available naturally url contains the object name
      return self.cfg_url
    # else get from stream config metadata
    return self.cfg_stream_config_metadata.get('DATASET_OBJECT_NAME')

  @property
  def parent_fld(self):
    return self.os_path.join(self.log.get_target_folder('data'), 'TRAINING_DATASETS')
  
  def _setup_minio(self):
    # TODO: check for self.cfg_stream_config_metadata for override minio config !
    if 'MINIO_CONFIG' in self.cfg_stream_config_metadata:
      self._custom_minio = self.cfg_stream_config_metadata['MINIO_CONFIG']
    else:
      self._file_system_manager = self.shmem['file_system_manager']
      if self._file_system_manager.file_system.signature.lower() != "minio":
        self.P("Errors will occur because the configured file system on this box is not minio!", color='e')
    return

  @property
  def ds_local_path(self):
    res = self.os_path.join(self.parent_fld, self.cfg_dataset_object_name)
    return res if not res.endswith('.zip') else res[:-4]

  def _list_minio_files(self):
    res = None
    if self._file_system_manager is not None:
      res = self._file_system_manager.file_system.list_objects()
    else:
      # TODO: implement
      pass
    return res
  
  def _download_minio_file(self, uri, local_file_path):
    res = None
    if self._file_system_manager is not None:
      res = self._file_system_manager.download(
        uri=uri,
        local_file_path=local_file_path
      )
    else:
      # TODO: implement
      pass
    return res

  def _init(self):
    return

  def _release(self):
    return

  def _maybe_reconnect(self):
    return

  def _run_data_aquisition_step(self):
    if not self.checked_dataset:
      self.P(f"Checking if dataset {self.cfg_dataset_object_name} is available...")
    if not self.checked_dataset and not self._is_dataset_available():
      self.P(f"Dataset {self.cfg_dataset_object_name} not available.")
      self.add_ds_inputs(ready=False)
    else:
      self.checked_dataset = True
      if not self._send_pings:
        self.process_found_dataset()
        self._send_pings = True
      else:
        self._add_ping_inputs()
      #endif
    #endif
    return

  def process_found_dataset(self):
    if self.os_path.isdir(self.ds_local_path) and not self.cfg_force_download:
      self.P(f"Dataset already available at {self.ds_local_path}")
      self.dataset_status = "Dataset available"
      self.add_ds_inputs(ready=True)
      return
    # endif dataset already available

    self.dataset_status = "Downloading dataset"
    os.makedirs(self.parent_fld, exist_ok=True)
    dataset_path = self.ds_local_path + '.zip'
    cloud_path = self.cfg_dataset_object_name
    if not cloud_path.endswith('.zip'):
      cloud_path += '.zip'
    self.P(f"Downloading dataset {cloud_path} to {dataset_path}")
    self._download_minio_file(
      uri=cloud_path,
      local_file_path=dataset_path
    )
    self.P(f"Downloaded dataset {self.cfg_dataset_object_name} to {dataset_path}. Unzipping...")
    self.dataset_status = "Unzipping dataset"
    target_path = self._unzip(dataset_path, self.ds_local_path)
    self.P(f"Unzipped dataset {self.cfg_dataset_object_name} to {target_path}.")
    os.remove(dataset_path)
    self.dataset_status = "Dataset available"
    self.add_ds_inputs(ready=True)
    return

  def add_ds_inputs(self, ready=False):
    self._add_inputs([
      self._new_input(struct_data={
        'dataset_ready': ready,
        'dataset_path': self.ds_local_path if ready else None,
        'dataset_status': self.dataset_status
      })
    ])

  def _add_ping_inputs(self):
    self._add_inputs(
      [
        self._new_input(struct_data={
          'ping': True,
          'dataset_ready': True,
          'dataset_status': self.dataset_status,
          'dataset_path': self.ds_local_path
        })
      ]
    )
    return

  def _is_dataset_available(self):
    target_path = self.os_path.join(self.parent_fld, self.cfg_dataset_object_name)
    if self.os_path.isdir(target_path) and not self.cfg_force_download:
      return True
    inventory = self._file_system_manager.file_system.list_objects()
    return self.cfg_dataset_object_name in inventory or self.cfg_dataset_object_name + '.zip' in inventory

  def _unzip(self, path_zip_file, dir_name=None):
    if dir_name is None:
      dir_name, ext = self.os_path.splitext(path_zip_file)
    with zipfile.ZipFile(path_zip_file, "r") as zip_ref:
      zip_ref.extractall(dir_name)
    return dir_name
